from pathlib import Path
import sys
sys.path.append('.')

from frads import ncp
from frads import parsers
from frads import geom




def test_gen_ports_from_window_ncp():
    """Generate sun sources for matrix generation."""
    window_polygon = geom.Polygon([geom.Vector(0, 0, 0),
                                   geom.Vector(0, 0, 4),
                                   geom.Vector(3, 0, 4),
                                   geom.Vector(3, 0, 0)])
    awning_polygon = [geom.Polygon([geom.Vector(0, -1, 2),
                                    geom.Vector(0, 0, 4),
                                    geom.Vector(3, 0, 4),
                                    geom.Vector(3, -1, 2)])]
    ports = ncp.gen_ports_from_window_ncp(window_polygon, awning_polygon)
    assert len(ports) == 5
    assert sum([p.area for p in ports]) == 26




from pathlib import Path
import os
import glob
import sys
sys.path.append(".")
from frads import epjson2rad



test_dir = Path(__file__).resolve().parent


def test_cfscase():
    epjson1 = test_dir / "Resources" / "CmplxGlz_SingleZone_DoubleClearAir.epJSON"
    epjson2rad.epjson2rad(epjson2rad.read_ep_input(epjson1))
    assert os.path.isfile("Room_102.cfg")
    assert os.path.isfile("./Objects/Room102_South_Wall.rad")
    assert os.path.isfile("./Objects/Room102_South_Wall_window.rad")
    assert os.path.isfile("./Objects/Room102_Ceiling.rad")
    os.remove("Room_102.cfg")
    obj_files = glob.glob("./Objects/Room_102*.rad")
    for file in obj_files:
        os.remove(file)

def test_refcase():
    epjson2 = test_dir / "Resources" / "RefBldgMediumOfficeNew2004_Chicago.epJSON"
    epjson2rad.epjson2rad(epjson2rad.read_ep_input(epjson2))
    assert os.path.isfile("./Objects/Perimeter_bot_ZN_1_Wall_South.rad")
    assert os.path.isfile("./Objects/Perimeter_mid_ZN_3_Ceiling.rad")
    assert os.path.isfile("Perimeter_bot_ZN_1.cfg")
    assert os.path.isfile("Perimeter_bot_ZN_2.cfg")
    assert os.path.isfile("Perimeter_bot_ZN_3.cfg")
    assert os.path.isfile("Perimeter_bot_ZN_4.cfg")
    assert os.path.isfile("Perimeter_mid_ZN_1.cfg")
    assert os.path.isfile("Perimeter_mid_ZN_2.cfg")
    assert os.path.isfile("Perimeter_mid_ZN_3.cfg")
    assert os.path.isfile("Perimeter_mid_ZN_4.cfg")
    assert os.path.isfile("Perimeter_top_ZN_1.cfg")
    assert os.path.isfile("Perimeter_top_ZN_2.cfg")
    assert os.path.isfile("Perimeter_top_ZN_3.cfg")
    assert os.path.isfile("Perimeter_top_ZN_4.cfg")
    cfg_files = glob.glob("./Perimeter*.cfg")
    obj_files = glob.glob("./Objects/Perimeter*.rad")
    for file in obj_files:
        os.remove(file)
    for file in cfg_files:
        os.remove(file)



"""Typical Radiance matrix-based simulation workflows
"""

from configparser import ConfigParser
from contextlib import contextmanager
import logging
import math
import os
from pathlib import Path
import shutil
import subprocess as sp
import tempfile as tf
from typing import List
from typing import Dict
from typing import Generator
from typing import Sequence
from typing import Tuple

from frads import geom
from frads import sky
from frads import matrix

from frads import mtxmult
from frads import parsers
from frads import raycall
from frads import utils
from frads.types import Options
from frads.types import Primitive
from frads.types import Receiver
from frads.types import Sender
from frads.types import MradModel
from frads.types import MradPath
from frads.types import View
from frads.types import WeaMetaData
from frads.types import WeaData


logger: logging.Logger = logging.getLogger("frads.methods")


def get_window_group(wpaths: List[Path]) -> Tuple[dict, list]:
    """Parse window groups from config.

    Args:
        wpaths(str): window file paths
    Return:
        window_groups(dict): window group name and primitives
        window_normals(list): a list of normal for each window group.
    """
    window_groups = {}
    window_normals: List[geom.Vector] = []
    for wpath in wpaths:
        prims = utils.unpack_primitives(wpath)
        window_groups[wpath.stem] = prims
        # Taking normal from the first polygon
        _normal = parsers.parse_polygon(prims[0].real_arg).normal
        if _normal not in window_normals:
            window_normals.append(_normal)
    return window_groups, window_normals


def get_ncp_shades(npaths: List[Path]) -> dict:
    """Parse ncp shade groups from config."""
    ncp_shades = {}
    for npath in npaths:
        prims = utils.unpack_primitives(npath)
        ncp_shades[npath.stem] = prims
    return ncp_shades


def get_wea_data(config: ConfigParser) -> Tuple[WeaMetaData, List[WeaData], str]:
    """Get wea data and parse into appropriate data types."""
    if wea_path := config["Site"].getpath("wea_path"):
        logger.info("Using user specified %s file.", wea_path)
        name = wea_path.stem
        with open(wea_path, "r", encoding="utf-8") as rdr:
            wea_metadata, wea_data = parsers.parse_wea(rdr.read())
    elif epw_path := config["Site"].getpath("epw_path"):
        logger.info("Converting %s to a .wea file", epw_path)
        name = epw_path.stem
        with open(epw_path, "r", encoding="utf-8") as rdr:
            wea_metadata, wea_data = parsers.parse_epw(rdr.read())
    else:
        raise ValueError("Need either a .wea or a .epw file")
    return wea_metadata, wea_data, name


def get_sender_grid(config: ConfigParser) -> Dict[str, Sender]:
    """Get point grid as ray senders."""
    sender_grid: Dict[str, Sender] = {}
    if (grid_paths := config["RaySender"].getpaths("grid_surface")) is None:
        return sender_grid
    for gpath in grid_paths:
        name: str = gpath.stem
        # Take the first polygon primitive
        gprimitives = utils.unpack_primitives(gpath)
        surface_polygon = None
        for prim in gprimitives:
            if prim.ptype == "polygon":
                surface_polygon = parsers.parse_polygon(prim.real_arg)
        if surface_polygon is None:
            raise ValueError(f"No polygon found in {gpath}")
        sensor_pts = utils.gen_grid(
            surface_polygon,
            config["RaySender"].getfloat("grid_height"),
            config["RaySender"].getfloat("grid_spacing"),
        )
        sender_grid[name] = matrix.points_as_sender(
            pts_list=sensor_pts, ray_cnt=config["SimControl"].getint("ray_count")
        )
    return sender_grid


def get_sender_view(config: ConfigParser) -> Tuple[dict, dict]:
    """Get a single view as a sender.
    Args:
        config: MradConfig object"""
    sender_view: Dict[str, matrix.Sender] = {}
    view_dicts: Dict[str, View] = {}
    if (view := config["RaySender"].getview("view")) is None:
        return sender_view, view_dicts
    view_name = "view_00"
    # if "vf" in vdict:
    # with open(vdict["vf"], "r", encoding="ascii") as rdr:
    # vdict.update(parsers.parse_vu(rdr.read()))
    view_dicts[view_name] = view
    sender_view[view_name] = matrix.view_as_sender(
        view=view,
        ray_cnt=int(config["SimControl"]["ray_count"]),
        xres=view.xres,
        yres=view.yres,
    )
    return sender_view, view_dicts


@contextmanager
def assemble_model(config: ConfigParser) -> Generator:
    """Assemble all the pieces together."""
    logger.info("Model assembling")
    # Get Ray senders
    sender_grid = get_sender_grid(config)
    sender_view, view_dicts = get_sender_view(config)
    if (not sender_grid) and (not sender_view):
        raise ValueError("Need to at least specify a grid or a view")
    # Get materials
    material_primitives = []
    for path in config["Model"].getpaths("material"):
        for prim in utils.unpack_primitives(path):
            material_primitives.append(prim)
    material_primitives.append(
        Primitive("void", "plastic", "black", ["0"], [5, 0, 0, 0, 0, 0])
    )
    material_primitives.append(
        Primitive("void", "glow", "glowing", ["0"], [4, 1, 1, 1, 0])
    )
    material_path = Path(f"all_material_{utils.id_generator()}.rad")
    with open(material_path, "w", encoding="ascii") as wtr:
        for primitive in material_primitives:
            wtr.write(str(primitive) + "\n")
    # Get window groups
    window_groups, window_normals = get_window_group(
        config["Model"].getpaths("windows", []), 
    )
    # Get BSDFs
    bsdf_mat = {
        wname: Path(path)
        for wname, path in zip(
            window_groups, config["Model"].getpaths("window_xmls", [])
        )
    }
    # Get ncp shades
    ncp_shades = get_ncp_shades(config["Model"].getpaths("ncps", []))
    # Get cfs paths
    cfs_path = config["Model"].getpaths("window_cfs", [])
    black_env_path = Path(f"blackened_{utils.id_generator()}.rad")
    with open(black_env_path, "w", encoding="ascii") as wtr:
        for path in config["Model"].getpaths("scene"):
            wtr.write(f"\n!xform -m black {path}")
    yield MradModel(
        config["Model"].get("name"),
        material_path,
        window_groups,
        window_normals,
        sender_grid,
        sender_view,
        view_dicts,
        bsdf_mat,
        cfs_path,
        ncp_shades,
        black_env_path,
    )
    logger.info("Cleaning up")
    os.remove(material_path)
    os.remove(black_env_path)


def prep_2phase_pt(mpath: MradPath, model: MradModel, config: ConfigParser) -> None:
    """Prepare matrices two phase methods."""
    logger.info("Computing for 2-phase sensor point matrices...")
    sys_paths = [
        model.material_path,
        *config["Model"].getpaths("scene"),
        *config["Model"].getpaths("windows"),
    ]
    opt = config["SimControl"].getoptions("dsmx_opt")
    opt["n"] = config["SimControl"].getint("nprocess")
    overwrite = config.getboolean("SimControl", "overwrite", fallback=False)
    for grid_name, sender_grid in model.sender_grid.items():
        mpath.pdsmx[grid_name] = Path("Matrices", f"pdsmx_{model.name}_{grid_name}.mtx")
        receiver_sky = matrix.sky_as_receiver(
            config["SimControl"]["smx_basis"],
            out=mpath.pdsmx[grid_name],
        )
        if (not mpath.pdsmx[grid_name].is_file()) or overwrite:
            matrix.rfluxmtx(sender_grid, receiver_sky, sys_paths, utils.opt2list(opt))


def prep_2phase_vu(mpath: MradPath, model: MradModel, config: ConfigParser) -> None:
    """Generate image-based matrices if view defined."""
    logger.info("Computing for image-based 2-phase matrices...")
    sys_paths = [
        model.material_path,
        *config["Model"].getpaths("scene"),
        *config["Model"].getpaths("windows"),
    ]
    opt = config["SimControl"].getoptions("dsmx_opt")
    opt["n"] = config["SimControl"].getint("nprocess")
    overwrite = config.getboolean("SimControl", "overwrite", fallback=False)
    for view_name, sender_view in model.sender_view.items():
        mpath.vdsmx[view_name] = Path(
            "Matrices", f"vdsmx_{model.name}_{view_name}", "%04d.hdr"
        )
        receiver_sky = matrix.sky_as_receiver(
            config["SimControl"]["smx_basis"],
            out=mpath.vdsmx[view_name],
        )
        if (not mpath.vdsmx[view_name].is_dir()) or overwrite:
            logger.info("Generating for %s", view_name)
            matrix.rfluxmtx(sender_view, receiver_sky, sys_paths, utils.opt2list(opt))


def view_matrix_pt(
    mpath: MradPath, model: MradModel, config: ConfigParser, direct: bool = False
) -> None:
    """."""
    _opt = config["SimControl"].getoptions("vmx_opt")
    _env: List[Path] = [
        model.material_path,
        *config["Model"].getpaths("scene"),
    ]
    if direct:
        logger.info("Computing direct view matrix for sensor grid:")
        _opt["ab"] = 1
        _env = [model.material_path, model.black_env_path]
    else:
        logger.info("Computing view matrix for sensor grid:")
    receiver_windows = Receiver(receiver="", basis=config["SimControl"]["vmx_basis"])
    for grid_name, sender_grid in model.sender_grid.items():
        for wname, window_prim in model.window_groups.items():
            _name = grid_name + wname
            if direct:
                mpath.pvmxd[_name] = Path(
                    "Matrices", f"pvmx_{model.name}_{_name}_d.mtx"
                )
                out = mpath.pvmxd[_name]
            else:
                mpath.pvmx[_name] = Path("Matrices", f"pvmx_{model.name}_{_name}.mtx")
                out = mpath.pvmx[_name]
            receiver_windows += matrix.surface_as_receiver(
                prim_list=window_prim,
                basis=config["SimControl"]["vmx_basis"],
                offset=None,
                left=None,
                source="glow",
                out=out,
            )
        if direct:
            files_exist = all(f.is_file() for f in mpath.pvmxd.values())
        else:
            files_exist = all(f.is_file() for f in mpath.pvmx.values())
        if (not files_exist) or config.getboolean(
            "SimControl", "overwrite", fallback=False
        ):
            logger.info("Generating vmx for %s", grid_name)
            matrix.rfluxmtx(sender_grid, receiver_windows, _env, utils.opt2list(_opt))


def view_matrix_vu(
    mpath: MradPath, model: MradModel, config: ConfigParser, direct: bool = False
) -> None:
    """Prepare matrices using three-phase methods."""
    _opt = config["SimControl"].getoptions("vmx_opt")
    _env = [model.material_path, *config["Model"].getpaths("scene")]
    direct_msg = ""
    if direct:
        _opt["i"] = True
        _opt["ab"] = 1
        _env = [model.material_path, model.black_env_path]
        direct_msg = " direct-only"
    _opt["n"] = config["SimControl"].getint("nprocess")
    overwrite = config.getboolean("SimControl", "overwrite", fallback=False)
    for view, sender_view in model.sender_view.items():
        vrcvr_windows = Receiver(receiver="", basis=config["SimControl"]["vmx_basis"])
        for wname, window_prim in model.window_groups.items():
            _name = view + wname
            if direct:
                mpath.vvmxd[_name] = Path(
                    "Matrices", f"vvmx_{model.name}_{_name}_d", "%04d.hdr"
                )
                out = mpath.vvmxd[_name]
            else:
                mpath.vvmx[_name] = Path(
                    "Matrices", f"vvmx_{model.name}_{_name}", "%04d.hdr"
                )
                out = mpath.vvmx[_name]
            out.parent.mkdir(exist_ok=True)
            vrcvr_windows += matrix.surface_as_receiver(
                prim_list=window_prim,
                basis=config["SimControl"]["vmx_basis"],
                source="glow",
                out=out,
            )
        if direct:
            exists = all(
                any(d.iterdir()) for d in mpath.vvmxd[_name].parents[1].glob("vvmx*_d")
            )
        else:
            exists = all(
                any(d.iterdir())
                for d in mpath.vvmx[_name].parents[1].glob("vvmx*[!_d]")
            )
        if (not exists) or overwrite:
            logger.info("Computing%s image-based view matrix", direct_msg)
            matrix.rfluxmtx(sender_view, vrcvr_windows, _env, utils.opt2list(_opt))


def daylight_matrix(
    mpath: MradPath, model: MradModel, config: ConfigParser, direct: bool = False
) -> None:
    """Call rfluxmtx to generate daylight matrices for each sender surface."""
    logger.info("Computing daylight matrix...")
    dmx_opt = config["SimControl"].getoptions("dmx_opt")
    dmx_opt["n"] = config["SimControl"].getint("nprocess")
    dmx_env = [model.material_path, *config["Model"].getpaths("scene")]
    if direct:
        dmx_opt["ab"] = 0
        dmx_env = [model.material_path, model.black_env_path]
    for sname, surface_primitives in model.window_groups.items():
        _name = sname
        if direct:
            mpath.dmxd[sname] = Path("Matrices", f"dmx_{model.name}_{_name}_d.mtx")
            out = mpath.dmxd[sname]
        else:
            mpath.dmx[sname] = Path("Matrices", f"dmx_{model.name}_{_name}.mtx")
            out = mpath.dmx[sname]
        if regen(out, config):
            logger.info("Generating daylight matrix for %s", _name)
            sndr_window = matrix.surface_as_sender(
                prim_list=surface_primitives, basis=config["SimControl"]["vmx_basis"]
            )
            receiver_sky = matrix.sky_as_receiver(
                config["SimControl"]["smx_basis"], out
            )
            matrix.rfluxmtx(sndr_window, receiver_sky, dmx_env, utils.opt2list(dmx_opt))


def blacken_env(model: MradModel, config: ConfigParser) -> Tuple[str, str]:
    """."""
    bwindow_path = f"blackened_window_{utils.id_generator()}.rad"
    gwindow_path = f"glowing_window_{utils.id_generator()}.rad"
    blackened_window = []
    glowing_window = []
    for _, windows in model.window_groups.items():
        for window in windows:
            blackened_window.append(
                Primitive(
                    "black",
                    window.ptype,
                    window.identifier,
                    window.str_arg,
                    window.real_arg,
                )
            )
            glowing_window.append(
                Primitive(
                    "glowing",
                    window.ptype,
                    window.identifier,
                    window.str_arg,
                    window.real_arg,
                )
            )
    with open(bwindow_path, "w", encoding="ascii") as wtr:
        wtr.write("\n".join(list(map(str, blackened_window))))
    with open(gwindow_path, "w", encoding="ascii") as wtr:
        wtr.write("\n".join(list(map(str, glowing_window))))
    vmap_oct = f"vmap_{utils.id_generator()}.oct"
    cdmap_oct = f"cdmap_{utils.id_generator()}.oct"
    raycall.oconv(
        str(model.material_path),
        *map(str, config["Model"].getpaths("scene")),
        gwindow_path,
        outpath=vmap_oct,
        frozen=True,
    )
    logger.info("Generating view matrix material map octree")
    raycall.oconv(
        str(model.material_path),
        *map(str, config["Model"]["scene"].split()),
        bwindow_path,
        outpath=cdmap_oct,
        frozen=True,
    )
    logger.info("Generating direct-sun matrix material map octree")
    os.remove(bwindow_path)
    os.remove(gwindow_path)
    return vmap_oct, cdmap_oct


def direct_sun_matrix_pt(
    mpath: MradPath, model: MradModel, config: ConfigParser
) -> None:
    """Compute direct sun matrix for sensor points.
    Args:
        smx_path: path to sun only sky matrix
    Returns:
        path to resulting direct sun matrix
    """

    logger.info("Direct sun matrix for sensor grid")
    cdsenv = [model.material_path, model.black_env_path, *model.cfs_paths]
    _cfs_name = "".join([Path(cfs).stem for cfs in model.cfs_paths])
    for grid_name, sender_grid in model.sender_grid.items():
        mpath.pcdsmx[grid_name] = Path(
            "Matrices", f"pcdsmx_{model.name}_{grid_name}_{_cfs_name}.mtx"
        )
        if regen(mpath.pcdsmx[grid_name], config):
            logger.info("Generating using rcontrib...")
            rcvr_sun = matrix.sun_as_receiver(
                basis="r6",
                smx_path=mpath.smx_sun,
                window_normals=model.window_normals,
                full_mod=True,
            )
            cdsmx_opt = config["SimControl"].getoptions("cdsmx_opt")
            cdsmx_opt["n"] = config["SimControl"].getint("nprocess")
            sun_oct = Path(f"sun_{utils.id_generator()}.oct")
            matrix.rcvr_oct(rcvr_sun, cdsenv, sun_oct)
            matrix.rcontrib(
                sender_grid,
                rcvr_sun.modifier,
                sun_oct,
                mpath.pcdsmx[grid_name],
                utils.opt2list(cdsmx_opt),
            )
            sun_oct.unlink()


def direct_sun_matrix_vu(
    mpath: MradPath, model: MradModel, vmap_oct, cdmap_oct, config: ConfigParser
) -> None:
    """Compute direct sun matrix for images.
    Args:
        mpath:
        model:
        vmap_oct:
        cdmap_oct:
        config:
    Returns:
        None
    """
    logger.info("Direct sun matrix for view (image)")
    rcvr_sun = matrix.sun_as_receiver(
        basis="r6", smx_path=mpath.smx_sun_img, window_normals=model.window_normals
    )
    mod_names = [f"{int(line[3:]):04d}" for line in rcvr_sun.modifier.splitlines()]
    sun_oct = Path(f"sun_{utils.id_generator()}.oct")
    cdsenv = [model.material_path, model.black_env_path, *model.cfs_paths]
    _cfs_name = "".join([Path(cfs).stem for cfs in model.cfs_paths])
    matrix.rcvr_oct(rcvr_sun, cdsenv, sun_oct)
    cdsmx_opt = config["SimControl"].getoptions("cdsmx_opt")
    cdsmx_opt["n"] = config["SimControl"].getint("nprocess")
    cdsmx_opt_list = utils.opt2list(cdsmx_opt)
    for view, sndr in model.sender_view.items():
        mpath.vmap[view] = Path("Matrices", f"vmap_{model.name}_{view}.hdr")
        mpath.cdmap[view] = Path("Matrices", f"cdmap_{model.name}_{view}.hdr")
        rpict_opt = Options()
        rpict_opt.ps = 1
        rpict_opt.ab = 0
        rpict_opt.av = (0.31831, 0.31831, 0.31831)
        cmd = raycall.get_rpict_command(model.views[view], rpict_opt, octree=vmap_oct)
        logger.info("Generating view matrix material map with: \n %s", " ".join(cmd))
        utils.run_write(cmd, mpath.vmap[view])
        cmd[-1] = cdmap_oct
        logger.info(
            "Generating direct-sun matrix material map with: \n %s", " ".join(cmd)
        )
        utils.run_write(cmd, mpath.cdmap[view])
        mpath.vcdfmx[view] = Path("Matrices", f"vcdfmx_{model.name}_{view}_{_cfs_name}")
        mpath.vcdrmx[view] = Path("Matrices", f"vcdrmx_{model.name}_{view}_{_cfs_name}")
        tempf = Path("Matrices", "vcdfmx")
        tempr = Path("Matrices", "vcdrmx")
        if regen(mpath.vcdfmx[view], config):
            logger.info("Generating direct sun f matrix for %s", view)
            matrix.rcontrib(sndr, rcvr_sun.modifier, sun_oct, tempf, cdsmx_opt_list)
            mpath.vcdfmx[view].mkdir(exist_ok=True)
            for idx, file in enumerate(sorted(tempf.glob("*.hdr"))):
                file.replace(mpath.vcdfmx[view] / (mod_names[idx] + ".hdr"))
            shutil.rmtree(tempf)
        if regen(mpath.vcdrmx[view], config):
            logger.info("Generating direct sun r matrix for %s", view)
            cdsmx_opt_list.append("-i+")
            matrix.rcontrib(sndr, rcvr_sun.modifier, sun_oct, tempr, cdsmx_opt_list)
            mpath.vcdrmx[view].mkdir(exist_ok=True)
            for idx, file in enumerate(sorted(tempr.glob("*.hdr"))):
                file.replace(mpath.vcdrmx[view] / (mod_names[idx] + ".hdr"))
            shutil.rmtree(tempr)
    sun_oct.unlink()


def calc_2phase_pt(
    mpath: MradPath,
    model: MradModel,
    datetime_stamps: Sequence[str],
) -> None:
    """."""
    logger.info("Computing for 2-phase sensor grid results.")
    for grid_name in mpath.pdsmx:
        grid_lines = model.sender_grid[grid_name].sender.decode().strip().splitlines()
        # we don't care about the direction part
        xypos = [",".join(line.split()[:3]) for line in grid_lines]
        opath = Path("Results", f"grid_{model.name}_{grid_name}.txt")
        res = mtxmult.mtxmult(mpath.pdsmx[grid_name], mpath.smx)
        if isinstance(res, bytes):
            res = res.decode().splitlines()
        else:
            res = ["\t".join(map(str, row)) for row in res.T.tolist()]
        with open(opath, "w", encoding="utf-8") as wtr:
            wtr.write("\t" + "\t".join(xypos) + "\n")
            for idx, value in enumerate(res):
                wtr.write(datetime_stamps[idx] + "\t")
                wtr.write(value.rstrip() + "\n")


def calc_2phase_vu(mpath: MradPath, model: MradModel, datetime_stamps) -> None:
    """."""
    logger.info("Computing for 2-phase image-based results")
    for view in mpath.vdsmx:
        opath = Path("Results", f"view_{model.name}_{view}")
        if opath.is_dir():
            shutil.rmtree(opath)
        cmd = mtxmult.get_imgmult_cmd(mpath.vdsmx[view], mpath.smx, odir=opath)
        logger.info(" ".join(cmd))
        sp.run(cmd, check=True)
        ofiles = sorted(opath.glob("*.hdr"))
        for idx, val in enumerate(ofiles):
            val.replace(opath / (datetime_stamps[idx] + ".hdr"))


def calc_3phase_pt(
    mpath: MradPath,
    model: MradModel,
    datetime_stamps: list,
) -> None:
    """."""
    logger.info("Computing for 3-phase sensor grid results")
    for grid_name in model.sender_grid:
        presl = []
        grid_lines = model.sender_grid[grid_name].sender.decode().strip().splitlines()
        xyzpos = [",".join(line.split()[:3]) for line in grid_lines]
        for wname in model.window_groups:
            _res = mtxmult.mtxmult(
                mpath.pvmx[grid_name + wname],
                model.bsdf_xml[wname],
                mpath.dmx[wname],
                mpath.smx,
            )
            if isinstance(_res, bytes):
                presl.append(
                    [
                        map(float, line.decode().strip().split("\t"))
                        for line in _res.splitlines()
                    ]
                )
            else:
                presl.append(_res.T.tolist())
        res = [[sum(tup) for tup in zip(*line)] for line in zip(*presl)]
        respath = Path("Results", f"grid_{model.name}_{grid_name}.txt")
        with open(respath, "w", encoding="utf-8") as wtr:
            wtr.write("\t" + "\t".join(xyzpos) + "\n")
            for idx, val in enumerate(res):
                wtr.write(datetime_stamps[idx] + "\t")
                wtr.write("\t".join(map(str, val)) + "\n")


def calc_3phase_vu(
    mpath: MradPath, model: MradModel, datetime_stamps, config: ConfigParser
) -> None:
    """."""
    for view in model.sender_view:
        opath = Path("Results", f"view_{model.name}_{view}")
        if opath.is_dir():
            shutil.rmtree(opath)
        logger.info("Computing for 3-phase image-based results for %s", view)
        vresl = []
        for wname in model.window_groups:
            _vrespath = Path("Results", f"{view}_{model.name}_{wname}")
            _vrespath.mkdir(exist_ok=True)
            cmd = mtxmult.get_imgmult_cmd(
                mpath.vvmx[view + wname],
                model.bsdf_xml[wname],
                mpath.dmx[wname],
                mpath.smx,
                odir=_vrespath,
            )
            logger.info(" ".join(cmd))
            sp.run(cmd, check=True)
            vresl.append(_vrespath)
        if len(vresl) > 1:
            ops = ["+"] * (len(vresl) - 1)
            mtxmult.batch_pcomb(
                vresl, ops, opath, nproc=config.getint("SimControl", "nprocess")
            )
            for path in vresl:
                shutil.rmtree(path)
        else:
            shutil.move(vresl[0], opath)
        ofiles = sorted(opath.glob("*.hdr"))
        for idx, ofile in enumerate(ofiles):
            ofile.replace(opath / (datetime_stamps[idx] + ".hdr"))


def calc_5phase_pt(
    mpath: MradPath,
    model: MradModel,
    datetime_stamps: Sequence[str],
) -> None:
    """."""
    logger.info("Computing sensor grid results")
    for grid_name in model.sender_grid:
        presl = []
        pdresl = []
        grid_lines = model.sender_grid[grid_name].sender.decode().strip().splitlines()
        xyzpos = [",".join(line.split()[:3]) for line in grid_lines]
        mult_cds = mtxmult.mtxmult(mpath.pcdsmx[grid_name], mpath.smx_sun)
        if isinstance(mult_cds, bytes):
            prescd = [
                list(map(float, line.decode().strip().split("\t")))
                for line in mult_cds.splitlines()
            ]
        else:
            prescd = mult_cds.T.tolist()
        for wname in model.window_groups:
            _res = mtxmult.mtxmult(
                mpath.pvmx[grid_name + wname],
                model.bsdf_xml[wname],
                mpath.dmx[wname],
                mpath.smx,
            )
            _resd = mtxmult.mtxmult(
                mpath.pvmxd[grid_name + wname],
                model.bsdf_xml[wname],
                mpath.dmxd[wname],
                mpath.smxd,
            )
            if isinstance(_res, bytes):
                _res = [
                    map(float, line.decode().strip().split("\t"))
                    for line in _res.splitlines()
                ]
                _resd = [
                    map(float, line.decode().strip().split("\t"))
                    for line in _resd.splitlines()
                ]
            else:
                _res = _res.T.tolist()
                _resd = _resd.T.tolist()
            presl.append(_res)
            pdresl.append(_resd)
        pres3 = [[sum(tup) for tup in zip(*line)] for line in zip(*presl)]
        pres3d = [[sum(tup) for tup in zip(*line)] for line in zip(*pdresl)]
        res = [
            [x - y + z for x, y, z in zip(a, b, c)]
            for a, b, c in zip(pres3, pres3d, prescd)
        ]
        respath = Path("Results", f"grid_{model.name}_{grid_name}.txt")
        with open(respath, "w", encoding="utf-8") as wtr:
            wtr.write("\t" + "\t".join(xyzpos) + "\n")
            for idx, val in enumerate(res):
                wtr.write(datetime_stamps[idx] + "\t")
                wtr.write("\t".join(map(str, val)) + "\n")


def calc_5phase_vu(
    mpath: MradPath,
    model: MradModel,
    datetime_stamps,
    datetime_stamps_d6,
    config: ConfigParser,
) -> None:
    """Compute for image-based 5-phase method result."""
    nprocess = config.getint("SimControl", "nprocess")
    for view in model.sender_view:
        logger.info("Computing for image-based results for %s", view)
        vresl = []
        vdresl = []
        with tf.TemporaryDirectory() as td:
            vrescdr = Path(tf.mkdtemp(dir=td))
            vrescdf = Path(tf.mkdtemp(dir=td))
            vrescd = Path(tf.mkdtemp(dir=td))
            cmds = []
            if mpath.vcdrmx != {}:
                cmds.append(
                    mtxmult.get_imgmult_cmd(
                        mpath.vcdrmx[view] / "%04d.hdr", mpath.smx_sun_img, odir=vrescdr
                    )
                )
            if mpath.vcdfmx != {}:
                cmds.append(
                    mtxmult.get_imgmult_cmd(
                        mpath.vcdfmx[view] / "%04d.hdr", mpath.smx_sun_img, odir=vrescdf
                    )
                )
            for wname in model.window_groups:
                _vrespath = tf.mkdtemp(dir=td)
                _vdrespath = tf.mkdtemp(dir=td)
                cmds.append(
                    mtxmult.get_imgmult_cmd(
                        mpath.vvmx[view + wname],
                        model.bsdf_xml[wname],
                        mpath.dmx[wname],
                        mpath.smx,
                        odir=Path(_vrespath),
                    )
                )
                cmds.append(
                    mtxmult.get_imgmult_cmd(
                        mpath.vvmxd[view + wname],
                        model.bsdf_xml[wname],
                        mpath.dmxd[wname],
                        mpath.smxd,
                        odir=Path(_vdrespath),
                    )
                )
                vresl.append(Path(_vrespath))
                vdresl.append(Path(_vdrespath))
            logger.info("Multiplying matrices for images.")
            for cmd in cmds:
                logger.info(" ".join(cmd))
            utils.batch_process(cmds, nproc=nprocess)
            logger.info("Combine results for each window groups.")
            res3 = Path(tf.mkdtemp(dir=td))
            res3di = Path(tf.mkdtemp(dir=td))
            res3d = Path(tf.mkdtemp(dir=td))
            if len(model.window_groups) > 1:
                ops = ["+"] * (len(vresl) - 1)
                mtxmult.batch_pcomb(vresl, ops, res3, nproc=nprocess)
                mtxmult.batch_pcomb(vdresl, ops, res3di, nproc=nprocess)
            else:
                for file in vresl[0].glob("*.hdr"):
                    file.replace(res3 / file.name)
                for file in vdresl[0].glob("*.hdr"):
                    file.replace(res3di / file.name)
            logger.info("Applying material reflectance map")
            mtxmult.batch_pcomb(
                [res3di, mpath.vmap[view]], ["*"], res3d, nproc=nprocess
            )
            mtxmult.batch_pcomb(
                [vrescdr, mpath.cdmap[view], vrescdf],
                ["*", "+"],
                vrescd,
                nproc=nprocess,
            )
            opath = Path("Results", f"view_{model.name}_{view}")
            if opath.is_dir():
                shutil.rmtree(opath)
            logger.info("Assemble all phase results.")
            res3_path = sorted(res3.glob("*.hdr"))
            for idx, stamp in enumerate(datetime_stamps):
                res3_path[idx].replace(res3 / (stamp + ".hdr"))
            res3d_path = sorted(res3d.glob("*.hdr"))
            for idx, stamp in enumerate(datetime_stamps):
                res3d_path[idx].replace(res3d / (stamp + ".hdr"))
            vrescd_path = sorted(vrescd.glob("*.hdr"))
            for idx, stamp in enumerate(datetime_stamps_d6):
                vrescd_path[idx].replace(vrescd / (stamp + ".hdr"))
            opath.mkdir(exist_ok=True)
            cmds = []
            opaths = []
            for hdr3 in os.listdir(res3):
                if hdr3 in os.listdir(vrescd):
                    opaths.append(opath / hdr3)
                    cmds.append(
                        [
                            "pcomb",
                            "-o",
                            str(res3 / hdr3),
                            "-s",
                            "-1",
                            "-o",
                            str(res3d / hdr3),
                            "-o",
                            str(vrescd / hdr3),
                        ]
                    )
                else:
                    os.replace(res3 / hdr3, opath / hdr3)
            if len(cmds) > 0:
                utils.batch_process(cmds, opaths=opaths)
            logger.info("Done computing for %s", view)


def regen(path: Path, config) -> bool:
    """
    Decides whether to regenerate a file depending on
    if the file already exists and if overwrite is on.
    """
    if path.is_file():
        exist = True
    elif path.is_dir():
        exist = True
    else:
        exist = False
    return (not exist) or config.getboolean("SimControl", "overwrite", fallback=False)


def two_phase(model: MradModel, config: ConfigParser) -> MradPath:
    """Two-phase simulation workflow."""
    mpath = MradPath()
    wea_meta, wea_data, wea_name = get_wea_data(config)
    mpath.smx = Path("Matrices") / (wea_name + ".smx")
    wea_data, datetime_stamps = sky.filter_wea(
        wea_data,
        wea_meta,
        daylight_hours_only=config.getboolean("Site", "daylight_hours_only"),
        start_hour=config.getfloat("Site", "start_hour"),
        end_hour=config.getfloat("Site", "end_hour"),
    )
    if regen(mpath.smx, config):
        sky.gendaymtx(
            mpath.smx,
            int(config["SimControl"]["smx_basis"][-1]),
            data=wea_data,
            meta=wea_meta,
            rotate=config["Site"].getfloat("orientation"),
        )
    prep_2phase_pt(mpath, model, config)
    prep_2phase_vu(mpath, model, config)
    if not config.getboolean("SimControl", "no_multiply", fallback=False):
        calc_2phase_pt(mpath, model, datetime_stamps)
        calc_2phase_vu(mpath, model, datetime_stamps)
    return mpath


def three_phase(
    model: MradModel, config: ConfigParser, direct: bool = False
) -> MradPath:
    """3/5-phase simulation workflow."""
    do_multiply = config.getboolean("SimControl", "no_multiply", fallback=False)
    psteps = 12 if direct else 4
    if do_multiply:
        psteps += 2
    mpath = MradPath()
    wea_meta, wea_data, wea_name = get_wea_data(config)
    mpath.smx = Path("Matrices") / (wea_name + ".smx")
    wea_data, datetime_stamps = sky.filter_wea(
        wea_data,
        wea_meta,
        daylight_hours_only=config.getboolean("Site", "daylight_hours_only"),
        start_hour=config.getfloat("Site", "start_hour"),
        end_hour=config.getfloat("Site", "end_hour"),
    )
    if regen(mpath.smx, config):
        sky.gendaymtx(
            mpath.smx,
            int(config["SimControl"]["smx_basis"][-1]),
            data=wea_data,
            meta=wea_meta,
            rotate=config["Site"].getfloat("orientation"),
        )
    view_matrix_pt(mpath, model, config)
    view_matrix_vu(mpath, model, config)
    daylight_matrix(mpath, model, config)
    if direct:
        if not (orientation := config["Site"].getfloat("orientation")) in (None, 0):
            rotate_radians = math.radians(orientation)
            rotated_window_normals = [
                n.rotate_3d(geom.Vector(0, 0, 1), rotate_radians) for n in model.window_normals
            ]
        wea_data_d6, datetime_stamps_d6 = sky.filter_wea(
            wea_data,
            wea_meta,
            daylight_hours_only=False,
            start_hour=0,
            end_hour=0,
            remove_zero=True,
            window_normals=rotated_window_normals if orientation else model.window_normals,
        )
        mpath.smxd = Path("Matrices") / (wea_name + "_d.smx")
        sky.gendaymtx(
            mpath.smxd,
            int(config["SimControl"]["smx_basis"][-1]),
            data=wea_data,
            meta=wea_meta,
            direct=True,
        )
        mpath.smx_sun_img = Path("Matrices") / (wea_name + "_d6_img.smx")
        sky.gendaymtx(
            mpath.smx_sun_img,
            int(config["SimControl"]["cdsmx_basis"][-1]),
            data=wea_data_d6,
            meta=wea_meta,
            rotate=config["Site"].getfloat("orientation"),
            onesun=True,
            direct=True,
        )
        mpath.smx_sun = Path("Matrices") / (wea_name + "_d6.smx")
        sky.gendaymtx(
            mpath.smx_sun,
            int(config["SimControl"]["cdsmx_basis"][-1]),
            data=wea_data,
            meta=wea_meta,
            rotate=config["Site"].getfloat("orientation"),
            onesun=True,
            direct=True,
        )
        vmap_oct, cdmap_oct = blacken_env(model, config)
        direct_sun_matrix_pt(mpath, model, config)
        if len(datetime_stamps_d6) > 0:
            direct_sun_matrix_vu(mpath, model, vmap_oct, cdmap_oct, config)
        daylight_matrix(mpath, model, config, direct=True)
        view_matrix_pt(mpath, model, config, direct=True)
        view_matrix_vu(mpath, model, config, direct=True)
        os.remove(vmap_oct)
        os.remove(cdmap_oct)
        if not do_multiply:
            calc_5phase_pt(
                mpath,
                model,
                datetime_stamps,
            )
            calc_5phase_vu(
                mpath,
                model,
                datetime_stamps,
                datetime_stamps_d6,
                config,
            )
    else:
        if not do_multiply:
            calc_3phase_pt(mpath, model, datetime_stamps)
            calc_3phase_vu(mpath, model, datetime_stamps, config)
    return mpath



"""Convert an EnergyPlus epJSON file into Radiance model[s]."""

from configparser import ConfigParser
import json
import logging
import math
import os
from pathlib import Path
import subprocess as sp
import sys
from typing import Dict, List

from frads import geom
from frads import utils
from frads.types import Primitive
from frads.types import BSDFData
from frads.types import RadMatrix
from frads.types import EPlusWindowGas
from frads.types import EPlusOpaqueMaterial
from frads.types import EPlusWindowMaterial
from frads.types import EPlusConstruction
from frads.types import EPlusOpaqueSurface
from frads.types import EPlusFenestration
from frads.types import EPlusZone

logger: logging.Logger = logging.getLogger("frads.epjson2rad")


def thicken(
    surface: geom.Polygon, windows: List[geom.Polygon], thickness: float
) -> list:
    """Thicken window-wall."""
    direction = surface.normal.scale(thickness)
    facade = surface.extrude(direction)[:2]
    for window in windows:
        facade.extend(window.extrude(direction)[2:])
    uniq = facade.copy()
    for idx, val in enumerate(facade):
        for rep in facade[:idx] + facade[idx + 1 :]:
            if set(val.to_list()) == set(rep.to_list()):
                uniq.remove(rep)
    return uniq


def get_construction_thickness(
    construction: EPlusConstruction, materials: dict
) -> float:
    """Get construction total thickness."""
    layer_thickness = []
    for layer in construction.layers:
        layer_thickness.append(materials[layer.lower()].thickness)
    return sum(layer_thickness)


def check_outward(polygon: geom.Polygon, zone_center: geom.Vector) -> bool:
    """Check whether a surface is facing outside."""
    outward = True
    angle2center = polygon.normal.angle_from(zone_center - polygon.centroid)
    if angle2center < math.pi / 4:
        outward = False
    return outward


def eplus_surface2primitive(
    surfaces: dict, constructions, zone_center, materials
) -> dict:
    """Conert EPlusOpaqueSurface (and its windows) to Radiance primitives."""
    surface_primitives: dict = {}
    for _, surface in surfaces.items():
        name = surface.name
        surface_primitives[name] = {}
        surface_primitives[name]["window"] = []
        surface_primitives[name]["xml"] = []
        surface_polygon = surface.polygon
        if not check_outward(surface_polygon, zone_center):
            surface_polygon = surface_polygon.flip()
        window_polygons = []
        for fen in surface.fenestrations:
            if constructions[fen.construction].type == "cfs":
                window_material = "void"
                surface_primitives[name]["xml"].append(fen.construction + ".xml")
            else:
                # Get the last construction layer
                window_material = (
                    constructions[fen.construction].layers[0].replace(" ", "_")
                )
            window_polygon = fen.polygon
            if check_outward(fen.polygon, zone_center):
                window_polygon = window_polygon.flip()
            window_polygons.append(window_polygon)
            surface_primitives[name]["window"].append(
                utils.polygon2prim(window_polygon, window_material, fen.name)
            )
        surface_primitives[name]["surface"] = []
        surface_construction_layers = constructions[surface.construction].layers
        # Second to the last of construction layers
        inner_material = surface_construction_layers[
            -min(2, len(surface_construction_layers))
        ].replace(" ", "_")
        surface_primitives[name]["surface"].append(
            utils.polygon2prim(surface_polygon, inner_material, surface.name)
        )
        if surface.sun_exposed:
            outer_material = surface_construction_layers[-1].replace(" ", "_")
            thickness = get_construction_thickness(
                constructions[surface.construction], materials
            )
            facade = thicken(surface_polygon, window_polygons, thickness)
            surface_primitives[name]["surface"].append(
                utils.polygon2prim(facade[1], outer_material, f"ext_{surface.name}")
            )
            for idx in range(2, len(facade)):
                surface_primitives[name]["surface"].append(
                    utils.polygon2prim(
                        facade[idx], inner_material, f"sill_{surface.name}.{idx}"
                    )
                )
    return surface_primitives


def write_primitives(surfaces: dict, odir: str) -> None:
    """Write surface and subsurface primitives."""
    for name, item in surfaces.items():
        opath = os.path.join(odir, name + ".rad")
        with open(opath, "w", encoding="ascii") as wtr:
            for primitive in item["surface"]:
                wtr.write(str(primitive))
        if item["window"] != []:
            opath = os.path.join(odir, name + "_window.rad")
            with open(opath, "w", encoding="ascii") as wtr:
                for primitive in item["window"]:
                    wtr.write(str(primitive))


def epluszone2rad(zone, constructions, materials):
    """Convert a EPlusZone object to a Radiance model."""
    zone_center = geom.polygon_center(*[wall.polygon for wall in zone.wall.values()])
    wall = eplus_surface2primitive(zone.wall, constructions, zone_center, materials)
    ceiling = eplus_surface2primitive(
        zone.ceiling, constructions, zone_center, materials
    )
    roof = eplus_surface2primitive(zone.roof, constructions, zone_center, materials)
    floor = eplus_surface2primitive(zone.floor, constructions, zone_center, materials)
    return wall, ceiling, roof, floor


def parse_material(name: str, material: dict) -> EPlusOpaqueMaterial:
    """Parser EP Material."""
    name = name.replace(" ", "_")
    roughness = material.get("roughness", "Smooth")
    thickness = material.get("thickness", 0.0)
    solar_absorptance = material.get("solar_absorptance", 0.7)
    visible_absorptance = material.get("visible_absorptance", 0.7)
    visible_reflectance = round(1 - visible_absorptance, 2)
    primitive = Primitive(
        "void",
        "plastic",
        name,
        ["0"],
        [5, visible_reflectance, visible_reflectance, visible_reflectance, 0, 0],
    )
    return EPlusOpaqueMaterial(
        name,
        roughness,
        solar_absorptance,
        visible_absorptance,
        visible_reflectance,
        primitive,
        thickness,
    )


def parse_material_nomass(name: str, material: dict) -> EPlusOpaqueMaterial:
    """Parse EP Material:NoMass"""
    name = name.replace(" ", "_")
    roughness = material.get("roughness", "Smooth")
    solar_absorptance = material.get("solar_absorptance", 0.7)
    visible_absorptance = material.get("visible_absorptance", 0.7)
    visible_reflectance = round(1 - visible_absorptance, 2)
    primitive = Primitive(
        "void",
        "plastic",
        name,
        ["0"],
        [5, visible_reflectance, visible_reflectance, visible_reflectance, 0, 0],
    )
    return EPlusOpaqueMaterial(
        name,
        roughness,
        solar_absorptance,
        visible_absorptance,
        visible_reflectance,
        primitive,
    )


def parse_windowmaterial_gap(name: str, material: dict) -> EPlusWindowGas:
    """Parse EP WindowMaterial:Gap"""
    name = name.replace(" ", "_")
    thickness = material["thickness"]
    gas = material["gas_or_gas_mixture_"]
    return EPlusWindowGas(name, thickness, gas, [1])


def parse_windowmaterial_gas(name: str, material: dict) -> EPlusWindowGas:
    """Parse EP WindowMaterial:Gas"""
    name = name.replace(" ", "_")
    ptype = [material["gas_type"]]
    thickness = material["thickness"]
    percentage = [1]
    return EPlusWindowGas(name, thickness, ptype, percentage)


def parse_windowmaterial_gasmixture(name: str, material: dict) -> EPlusWindowGas:
    """Parse EP WindowMaterial:GasMixture"""
    name = name.replace(" ", "_")
    thickness = material["thickness"]
    gas = [material["Gas"]]
    percentage = [1]
    return EPlusWindowGas(name, thickness, gas, percentage)


def parse_windowmaterial_simpleglazingsystem(
    name: str, material: dict
) -> EPlusWindowMaterial:
    """Parse EP WindowMaterial:SimpleGlazingSystem"""
    identifier = name.replace(" ", "_")
    shgc = material["solar_heat_gain_coefficient"]
    tmit = material.get("visible_transmittance", shgc)
    tmis = utils.tmit2tmis(tmit)
    primitive = Primitive("void", "glass", identifier, ["0"], [3, tmis, tmis, tmis])
    return EPlusWindowMaterial(identifier, tmit, primitive)


def parse_windowmaterial_simpleglazing(
    name: str, material: dict
) -> EPlusWindowMaterial:
    """Parse EP WindowMaterial:Simpleglazing"""
    identifier = name.replace(" ", "_")
    tmit = material["visible_transmittance"]
    tmis = utils.tmit2tmis(tmit)
    primitive = Primitive("void", "glass", identifier, ["0"], [3, tmis, tmis, tmis])
    return EPlusWindowMaterial(identifier, tmit, primitive)


def parse_windowmaterial_glazing(name: str, material: dict) -> EPlusWindowMaterial:
    """Parse EP WindowMaterial:Glazing"""
    identifier = name.replace(" ", "_")
    if material["optical_data_type"].lower() == "bsdf":
        tmit = 1
    else:
        tmit = material["visible_transmittance_at_normal_incidence"]
    tmis = utils.tmit2tmis(tmit)
    primitive = Primitive("void", "glass", identifier, ["0"], [3, tmis, tmis, tmis])
    return EPlusWindowMaterial(identifier, tmit, primitive)


def parse_windowmaterial_blind(inp: dict) -> dict:
    """Parse EP WindowMaterial:Blind"""
    blind_prims = {}
    for key, val in inp.items():
        _id = key.replace(" ", "_")
        # back_beam_vis_refl = val['back_side_slat_beam_visible_reflectance']
        # back_diff_vis_refl = val['back_side_slat_diffuse_visible_reflectance']
        # front_beam_vis_refl = val['front_side_slat_beam_visible_reflectance']
        front_diff_vis_refl = val["front_side_slat_diffuse_visible_reflectance"]
        # slat_width = val['slat_width']
        # slat_thickness = val['slat_thickness']
        # slat_separation = val['slat_separation']
        # slat_angle = val['slat_angle']
        blind_prims[key] = Primitive(
            "void",
            "plastic",
            _id,
            ["0"],
            [5, front_diff_vis_refl, front_diff_vis_refl, front_diff_vis_refl, 0, 0],
        )
        # genblinds_cmd = f"genblinds {_id} {_id} {slat_width} 3
        # {20*slat_separation} {slat_angle}"
    return blind_prims


def parse_epjson_material(epjs: dict) -> dict:
    """Parse each material type."""
    materials = {}
    for key, value in epjs.items():
        if "material" in key.split(":")[0].lower():
            for name, material in value.items():
                tocall = globals()[f"parse_{key.replace(':', '_')}".lower()]
                materials[name.lower()] = tocall(name, material)
    return materials


def parse_construction_complexfenestrationstate(epjs):
    """Parser EP Construction:ComplexFenestrationState."""
    construction = epjs.get("Construction:ComplexFenestrationState", {})
    cfs = {}
    matrices = {}
    for key, val in construction.items():
        val["ctype"] = "cfs"
        tf_name = val["visible_optical_complex_front_transmittance_matrix_name"]
        tb_name = val["visible_optical_complex_back_transmittance_matrix_name"]
        tf_list = epjs["Matrix:TwoDimension"][tf_name]["values"]
        tb_list = epjs["Matrix:TwoDimension"][tb_name]["values"]
        ncolumn = epjs["Matrix:TwoDimension"][tf_name]["number_of_columns"]
        # if ncolumn < 145:
        # raise ValueError("BSDF resolution too low to take advantage of Radiance")
        tf_bsdf = [v["value"] for v in tf_list]
        tb_bsdf = [v["value"] for v in tb_list]
        tf = utils.bsdf2sdata(BSDFData(tf_bsdf, ncolumn, ncolumn))
        tb = utils.bsdf2sdata(BSDFData(tb_bsdf, ncolumn, ncolumn))
        matrices[key] = RadMatrix(tf, tb)
        cfs[key] = EPlusConstruction(key, "cfs", [])
    return cfs, matrices


def parse_construction(construction: dict) -> dict:
    """Parse EP Construction"""
    constructions = {}
    for cname, clayer in construction.items():
        layers = list(clayer.values())
        constructions[cname] = EPlusConstruction(cname, "default", layers)
    return constructions


def parse_opaque_surface(surfaces: dict, fenestrations: dict) -> dict:
    """Parse opaque surface to a EPlusOpaqueSurface object."""
    opaque_surfaces = {}
    for name, surface in surfaces.items():
        identifier = name.replace(" ", "_")
        fenes = [fen for fen in fenestrations.values() if fen.host == name]
        ptype = surface["surface_type"]
        polygon = geom.Polygon(
            [geom.Vector(*vertice.values()) for vertice in surface["vertices"]]
        )
        for fen in fenes:
            polygon -= fen.polygon
        construction = surface["construction_name"]
        boundary = surface["outside_boundary_condition"]
        sun_exposed = surface["sun_exposure"] == "SunExposed"
        zone = surface["zone_name"]
        opaque_surfaces[name] = EPlusOpaqueSurface(
            identifier, ptype, polygon, construction, boundary, sun_exposed, zone, fenes
        )
    return opaque_surfaces


def parse_epjson_fenestration(fenes: dict) -> Dict[str, EPlusFenestration]:
    """Parse fenestration dictionary to a EPlusFenestration object."""
    fenestrations = {}
    for name, fen in fenes.items():
        surface_type = fen["surface_type"]
        if surface_type != "Door":
            name = name.replace(" ", "_")
            host = fen["building_surface_name"]
            vertices = []
            for i in range(1, fen["number_of_vertices"] + 1):
                vertices.append(
                    geom.Vector(
                        fen[f"vertex_{i}_x_coordinate"],
                        fen[f"vertex_{i}_y_coordinate"],
                        fen[f"vertex_{i}_z_coordinate"],
                    )
                )
            polygon = geom.Polygon(vertices)
            construction = fen["construction_name"]
            fenestrations[name] = EPlusFenestration(
                name, surface_type, polygon, construction, host
            )
    return fenestrations


def parse_epjson(epjs: dict) -> tuple:
    """
    Convert EnergyPlus JSON objects into Radiance primitives.
    """
    # get site information
    site = list(epjs["Site:Location"].values())[0]

    # parse each fenestration
    fenes = epjs.get("FenestrationSurface:Detailed")
    if fenes is not None:
        fenestrations = parse_epjson_fenestration(fenes)
    else:
        logger.warning("No fenestration found in the model.")
        sys.exit(1)

    # Get all the fenestration hosting surfaces
    fene_hosts = {val.host for val in fenestrations.values()}

    # parse each opaque surface
    opaque_surfaces = parse_opaque_surface(
        epjs["BuildingSurface:Detailed"], fenestrations
    )

    # parse each construction
    constructions = parse_construction(epjs["Construction"])

    cfs, matrices = parse_construction_complexfenestrationstate(epjs)
    constructions.update(cfs)

    # parse materials
    materials = parse_epjson_material(epjs)

    # get exterior zones
    exterior_zones = [
        value.zone
        for key, value in opaque_surfaces.items()
        if (key in fene_hosts) and value.sun_exposed
    ]

    # get secondary zones, but we don't do anything with it yet.
    secondary_zones: dict = {}
    for key, value in opaque_surfaces.items():
        if (key in fene_hosts) and (value.zone not in exterior_zones):
            adjacent_zone = opaque_surfaces[value.boundary].zone
            if adjacent_zone in exterior_zones:
                secondary_zones[value.zone] = {}

    zones = {}
    # go through each exterior zone, update zone dictionary.
    for zname in exterior_zones:
        zone_name = zname.replace(" ", "_")
        surface_map: dict = {"Wall": {}, "Ceiling": {}, "Roof": {}, "Floor": {}}
        windows = {
            n: val
            for n, val in fenestrations.items()
            if opaque_surfaces[val.host].zone == zname
        }
        for name, surface in opaque_surfaces.items():
            if surface.zone == zname:
                surface_map[surface.type][name] = surface
        zones[zname] = EPlusZone(
            zone_name,
            surface_map["Wall"],
            surface_map["Ceiling"],
            surface_map["Roof"],
            surface_map["Floor"],
            windows,
        )
    return site, zones, constructions, materials, matrices


def epjson2rad(epjs: dict) -> None:
    """Command-line program to convert a energyplus model into a Radiance model."""
    # Setup file structure
    Path("Objects").mkdir(exist_ok=True)
    Path("Matrices").mkdir(exist_ok=True)

    site, zones, constructions, materials, matrices = parse_epjson(epjs)
    building_name = epjs["Building"].popitem()[0].replace(" ", "_")
    if len(matrices) > 0:
        Path("Resources").mkdir(exist_ok=True)

    # Write material file
    material_path = os.path.join("Objects", f"materials{building_name}.mat")
    with open(material_path, "w", encoding="ascii") as wtr:
        for material in materials.values():
            wtr.write(str(material.primitive))

    # Write matrix files to xml, if any
    xml_paths = {}
    for key, val in matrices.items():
        opath = os.path.join("Resources", key + ".xml")
        tf_path = os.path.join("Resources", key + "_tf.mtx")
        tb_path = os.path.join("Resources", key + "_tb.mtx")
        with open(tf_path, "w", encoding="ascii") as wtr:
            wtr.write(repr(val.tf))
        with open(tb_path, "w", encoding="ascii") as wtr:
            wtr.write(repr(val.tb))
        # basis = ''.join([word[0] for word in val.tf.basis.split()])
        basis = "".join(
            [word[0].lower() for word in utils.BASIS_DICT[str(val.tf.ncolumn)].split()]
        )
        cmd = ["wrapBSDF", "-f", "n=" + key, "-a", basis]
        cmd += ["-tf", tf_path, "-tb", tb_path, "-U"]
        wb_process = sp.run(cmd, check=True, stdout=sp.PIPE, stderr=sp.PIPE)
        with open(opath, "wb") as wtr:
            wtr.write(wb_process.stdout)
        xml_paths[key] = opath

    # For each zone write primitves to files and create a config file
    for name, zone in zones.items():
        mrad_config = ConfigParser(allow_no_value=False)
        mrad_config["SimControl"] = {
            "vmx_basis": "kf",
            "vmx_opt": "-ab 5 -ad 65536 -lw 1e-5",
            "fmx_basis": "kf",
            "smx_basis": "r4",
            "dmx_opt": "-ab 2 -ad 128 -c 5000",
            "dsmx_opt": "-ab 7 -ad 16384 -lw 5e-5",
            "cdsmx_opt": "-ab 1",
            "cdsmx_basis": "r6",
            "ray_count": "1",
            "nprocess": "1",
            "separate_direct": "False",
            "overwrite": "False",
            "method": "",
        }
        mrad_config["Site"] = {
            "wea_path": "",
            "zipcode": "",
            "latitude": site["latitude"],
            "longitude": site["longitude"],
            "start_hour": "",
            "end_hour": "",
            "daylight_hours_only": "True",
        }
        primitives = epluszone2rad(zone, constructions, materials)
        scene = []
        windows = []
        window_xmls = []
        floors = []
        for primitive in primitives:
            write_primitives(primitive, "Objects")
            for _name, item in primitive.items():
                if item["surface"] != []:
                    scene.append(os.path.join("Objects", _name + ".rad"))
                if item["window"] != []:
                    windows.append(os.path.join("Objects", _name + "_window.rad"))
                if item["xml"] != []:
                    window_xmls.extend(
                        [os.path.join("Resources", xml) for xml in item["xml"]]
                    )
        # Get floors
        for primitive in primitives[-1]:
            floors.append(os.path.join("Objects", primitive + ".rad"))

        mrad_config["Model"] = {
            "material": material_path,
            "scene": "\n".join(scene),
            "windows": " ".join(windows),
            "window_xml": " ".join(window_xmls),
            "ncp_shade": "",
        }
        mrad_config["RaySender"] = {
            "grid_surface": " ".join(floors),
            "grid_spacing": "1",
            "grid_height": "0.75",
        }
        with open(f"{name.replace(' ', '_')}.cfg", "w", encoding="utf-8") as wtr:
            mrad_config.write(wtr)


def read_ep_input(fpath: Path) -> dict:
    """Load and parse input file into a JSON object.
    If the input file is in .idf fomart, use command-line
    energyplus program to convert it to epJSON format
    Args:
        fpath: input file path
    Returns:
        epjs: JSON object as a Python dictionary
    """
    epjson_path: Path
    if fpath.suffix == ".idf":
        cmd = ["energyplus", "--convert-only", str(fpath)]
        sp.run(cmd, check=True)
        epjson_path = Path(fpath.with_suffix(".epJSON").name)
        if not epjson_path.is_file():
            raise FileNotFoundError(f"Converted {str(epjson_path)} not found.")
    elif fpath.suffix == ".epJSON":
        epjson_path = fpath
    else:
        raise Exception(f"Unknown file type {fpath}")
    with open(epjson_path) as rdr:
        epjs = json.load(rdr)
    return epjs



"""
Executive command-line program for Radiance matrix-based simulation.
"""

import argparse
import configparser
import json
import logging
import os
from pathlib import Path
import sys
import tempfile as tf
from typing import List

from frads import color
from frads import color_data
from frads import epjson2rad
from frads import matrix
from frads import methods
from frads import mtxmult
from frads import ncp
from frads import parsers
from frads import room
from frads import utils
from frads.types import NcpModel
from frads.types import PaneRGB
from frads.types import Receiver


logger: logging.Logger = logging.getLogger("frads")


def mrad_init(args: argparse.Namespace) -> None:
    """Initiate mrad operation.

    Args:
        args: argparse.Namespace
    Returns:
        None
    """

    def get_file_list(paths: List[Path], ext):
        file_list: List[str] = []
        for path in paths:
            if not path.is_dir():
                file_list.append(str(path))
            else:
                file_list.extend(sorted(path.glob(f"*.{ext}")))
        return file_list

    config = configparser.ConfigParser(allow_no_value=False)
    simcontrol = {
        "# vmx_opt": "",
        "# dmx_opt": "",
        "# smx_basis": "",
        "# overwrite": "",
        "# separate_direct": "",
    }
    site = {}
    model = {}
    raysender = {}
    if args.wea_path is not None:
        if not args.wea_path.is_file():
            raise FileNotFoundError(args.wea_path)
        site["wea_path"] = args.wea_path
    elif args.epw_path is not None:
        if not args.epw_path.is_file():
            raise FileNotFoundError(args.epw_path)
        site["epw_path"] = args.epw_path
    model["name"] = args.name
    if args.grid is not None:
        raysender["grid_surface"] = args.grid[0]
        raysender["grid_spacing"] = args.grid[1]
        raysender["grid_height"] = args.grid[2]
    material_list = get_file_list(args.material, "mat")
    object_list: List[str] = get_file_list(args.object, "rad")
    model["scene"] = "\n".join(object_list)
    model["material"] = "\n".join(material_list)
    if args.window is not None:
        window_list = get_file_list(args.window, "rad")
        model["windows"] = "\n".join(window_list)
    if args.bsdf is not None:
        xml_list = get_file_list(args.bsdf, "xml")
        if len(window_list) != len(xml_list):
            raise ValueError("Number of window and xml files not the same")
        model["window_xmls"] = "\n".join(xml_list)
    templ_config = {
        "SimControl": simcontrol,
        "Site": site,
        "Model": model,
        "RaySender": raysender,
    }
    config.read_dict(templ_config)
    with open(f"{args.name}.cfg", "w", encoding="utf-8") as rdr:
        config.write(rdr)


def mrad_run(args: argparse.Namespace) -> None:
    """Call mtxmethod to carry out the actual simulation."""
    config = parsers.parse_mrad_config(args.cfg)
    if config["Model"].get("name") is None:
        config["Model"]["name"] = args.cfg.stem
    Path("Matrices").mkdir(exist_ok=True)
    Path("Results").mkdir(exist_ok=True)
    with methods.assemble_model(config) as model:
        if method := config["SimControl"].get("method"):
            if method.startswith(("2", "two")):
                methods.two_phase(model, config)
            elif method.startswith(("3", "three")):
                methods.three_phase(model, config)
            elif method.startswith(("5", "five")):
                methods.three_phase(model, config, direct=True)
        else:
            # Use 3- or 5-phase methods if we have
            # window groups and bsdf defined
            if model.window_groups and model.bsdf_xml:
                if config.getboolean("SimControl", "separate_direct"):
                    logger.info("Using five-phase simulation")
                    methods.three_phase(model, config, direct=True)
                else:
                    logger.info("Using three-phase simulation")
                    methods.three_phase(model, config)
            else:
                logger.info("Using two-phase method")
                methods.two_phase(model, config)


def mrad() -> None:
    """mrad entry point: parse arugments for init and run subprograms."""
    parser = argparse.ArgumentParser(
        prog="mrad",
        description="Mrad is an executive program for Radiance matrix-based simulation methods.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparser = parser.add_subparsers()
    # Parse arguments for init subprogram
    parser_init = subparser.add_parser("init")
    parser_init.set_defaults(func=mrad_init)
    egroup = parser_init.add_mutually_exclusive_group(required=True)
    egroup.add_argument("-a", "--wea_path", metavar="wea_path", type=Path)
    egroup.add_argument("-e", "--epw_path", metavar="epw_path", type=Path)
    parser_init.add_argument("-n", "--name", metavar="model_name", default="default")
    parser_init.add_argument(
        "-o",
        "--object",
        required=True,
        metavar="object",
        nargs="+",
        type=Path,
        help="Objects to include, can include wildcards",
    )
    parser_init.add_argument(
        "-m",
        "--material",
        required=True,
        metavar="material",
        nargs="+",
        type=Path,
        help="Material files to include, can include wildcards",
    )
    parser_init.add_argument(
        "-w",
        "--window",
        nargs="+",
        metavar="window",
        type=Path,
        help="Window files to include, these are gonna get parsed into window groups",
    )
    parser_init.add_argument(
        "-x",
        "--bsdf",
        nargs="+",
        metavar="bsdf",
        type=Path,
        help="xml file paths to include for each window file",
    )
    parser_init.add_argument(
        "-g",
        "--grid",
        nargs=3,
        metavar=("surface path", "grid spacing", "grid height"),
        help="Grid file path, grid spacing and height",
    )
    parser_init.add_argument(
        "-u", "--view", metavar="view", help="Grid file path, grid spacing and height"
    )
    # Parse arguments for run subprogram
    parser_run = subparser.add_parser("run")
    parser_run.set_defaults(func=mrad_run)
    parser_run.add_argument("cfg", type=Path, help="configuration file path")
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help=(
            "Verbose mode: -v=Debug, -vv=Info, -vvv=Warning, "
            "-vvvv=Error, -vvvvv=Critical"
        ),
    )
    args = parser.parse_args()
    # Setup logger
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s", "%y-%m-%d %H:%M:%S"
    )
    console_handler = logging.StreamHandler()
    _level = args.verbose * 10
    logger.setLevel(_level)
    console_handler.setLevel(_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    if args.verbose > 1:
        sys.tracebacklimit = 0
    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)
    args.func(args)


def glaze(args) -> None:
    """Command-line program for generating BRTDfunc for glazing system."""
    if args.optics is not None:
        panes = [parsers.parse_optics(fpath) for fpath in args.optics]
    else:
        panes = []
        for item in args.igsdb:
            try:
                with open(item, "r", encoding="utf-8") as rdr:
                    json_obj = json.load(rdr)
            except FileNotFoundError as fnfe:
                raise fnfe
            panes.append(parsers.parse_igsdb_json(json_obj))
    pane_rgb = []
    coeffs = color_data.XYZ2RGB_RAD
    # coeffs = color.get_conversion_matrix(args.cspace)
    for pane in panes:
        wvls = sorted(set(color_data.CIE_XYZ_2).intersection(pane.wavelength))
        wvl_range = wvls[-1] - wvls[0]
        cie_xyz = color.get_interpolated_cie_xyz(wvls, args.observer)
        trans = [pane.transmittance[pane.wavelength.index(wvl)] for wvl in wvls]
        reflf = [pane.reflectance_front[pane.wavelength.index(wvl)] for wvl in wvls]
        reflb = [pane.reflectance_back[pane.wavelength.index(wvl)] for wvl in wvls]
        tf_x, tf_y, tf_z = color.spec2xyz(cie_xyz, trans, wvl_range)
        rf_x, rf_y, rf_z = color.spec2xyz(cie_xyz, reflf, wvl_range)
        rb_x, rb_y, rb_z = color.spec2xyz(cie_xyz, reflb, wvl_range)
        tf_rgb = color.xyz2rgb(tf_x, tf_y, tf_z, coeffs)
        rf_rgb = color.xyz2rgb(rf_x, rf_y, rf_z, coeffs)
        rb_rgb = color.xyz2rgb(rb_x, rb_y, rb_z, coeffs)
        if pane.coated_side == "front":
            coated_rgb = rf_rgb
            glass_rgb = rb_rgb
        else:
            coated_rgb = rb_rgb
            glass_rgb = rf_rgb
        pane_rgb.append(PaneRGB(pane, coated_rgb, glass_rgb, tf_rgb))
    print(utils.get_glazing_primitive(pane_rgb))


def gengrid(args) -> None:
    """Commandline program for generating a grid of sensor points."""
    prims = utils.unpack_primitives(args.surface)
    polygon_prims = [prim for prim in prims if prim.ptype == "polygon"]
    polygon = parsers.parse_polygon(polygon_prims[0].real_arg).flip()
    if args.op:
        polygon = polygon.flip()
    grid_list = utils.gen_grid(polygon, args.height, args.spacing)
    cleanedup = []
    for row in grid_list:
        new_row = []
        for val in row:
            if val.is_integer():
                new_row.append(int(val))
            else:
                if (rounded := round(val, 1)).is_integer():
                    new_row.append(int(rounded))
                else:
                    new_row.append(rounded)
        cleanedup.append(new_row)
    grid_str = "\n".join([" ".join(map(str, row)) for row in cleanedup])
    print(grid_str)


def epjson2rad_cmd() -> None:
    """Command-line interface to converting epjson to rad."""
    parser = argparse.ArgumentParser()
    parser.add_argument("fpath", type=Path)
    parser.add_argument("-run", action="store_true", default=False)
    args = parser.parse_args()
    epjs = epjson2rad.read_ep_input(args.fpath)
    if "FenestrationSurface:Detailed" not in epjs:
        raise ValueError("No windows found in this model")
    epjson2rad.epjson2rad(epjs)


def genmtx_pts_sky(args) -> None:
    """Generate a point to sky matrix."""
    with open(args.pts, "r", encoding="ascii") as rdr:
        pts = [line.split() for line in rdr.readlines()]
    sender = matrix.points_as_sender(pts_list=pts, ray_cnt=1)
    out = Path(f"{args.pts.stem}_{args.basis}sky.mtx")
    receiver = matrix.sky_as_receiver(args.basis, out)
    sys_paths = args.sys
    del args.pts, args.basis, args.sys, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_vu_sky(args) -> None:
    """Generate a view to sky matrix."""
    with open(args.vu, "r", encoding="ascii") as rdr:
        view = parsers.parse_vu(rdr.readlines()[-1])  # use the last view
    sender = matrix.view_as_sender(
        view,
        ray_cnt=1,
        xres=args.resolu[0],
        yres=args.resolu[1],
    )
    out = Path(f"{args.vu.stem}_{args.basis}sky")
    out.mkdir()
    receiver = matrix.sky_as_receiver(args.basis, out / "%04d.hdr")
    sys_paths = args.sys
    del args.vu, args.basis, args.sys, args.resolu, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_srf_sky(args) -> None:
    """Generate a surface to sky matrix."""
    sender = matrix.surface_as_sender(
        prim_list=utils.unpack_primitives(args.srf),
        basis=args.basis[0],
        offset=args.offset,
    )
    out = Path(f"{args.srf.stem}_{'_'.join(args.basis)}sky.mtx")
    receiver = matrix.sky_as_receiver(args.basis[1], out)
    sys_paths = args.sys
    del args.srf, args.basis, args.sys, args.offset, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_pts_srf(args) -> None:
    """Generate a point to surface matrix."""
    with open(args.pts, "r", encoding="ascii") as rdr:
        pts = [line.split() for line in rdr.readlines()]
    sender = matrix.points_as_sender(pts_list=pts, ray_cnt=1)
    rprims = utils.unpack_primitives(args.srf)
    modifiers = {prim.modifier for prim in rprims if prim.ptype in ("polygon", "ring")}
    receiver = Receiver(receiver="", basis=args.basis, modifier="")
    sys_paths = args.sys
    for mod in modifiers:
        _receiver = [
            prim
            for prim in rprims
            if prim.modifier == mod and prim.ptype in ("polygon", "ring")
        ]
        if _receiver != []:
            outpath = Path(f"{args.pts.stem}_{args.srf.stem}.mtx")
            receiver += matrix.surface_as_receiver(
                prim_list=_receiver,
                basis=args.basis,
                offset=args.offset,
                left=None,
                source="glow",
                out=outpath,
            )
    del args.pts, args.srf, args.sys, args.basis, args.offset, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_vu_srf(args) -> None:
    """Generate a view to surface matrix."""
    with open(args.vu, "r", encoding="ascii") as rdr:
        view = parsers.parse_vu(rdr.readlines()[-1])
    sender = matrix.view_as_sender(
        view,
        ray_cnt=1,
        xres=args.resolu[0],
        yres=args.resolu[1],
    )
    rprims = utils.unpack_primitives(args.srf)
    modifiers = {prim.modifier for prim in rprims if prim.ptype in ("polygon", "ring")}
    receiver = Receiver(receiver="", basis=args.basis, modifier="")
    sys_paths = args.sys
    for mod in modifiers:
        _receiver = [
            prim
            for prim in rprims
            if prim.modifier == mod and prim.ptype in ("polygon", "ring")
        ]
        if _receiver != []:
            outpath = Path(f"{args.vu.stem}_{args.srf.stem}")
            outpath.mkdir()
            receiver += matrix.surface_as_receiver(
                prim_list=_receiver,
                basis=args.basis,
                offset=args.offset,
                left=None,
                source="glow",
                out=outpath / "%04d.hdr",
            )
    del args.vu, args.srf, args.sys, args.basis, args.offset, args.resolu, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_srf_srf(args) -> None:
    """Generate a surface to surface matrix."""
    sender = matrix.surface_as_sender(
        prim_list=utils.unpack_primitives(args.ssrf),
        basis=args.basis[0],
        offset=args.offset[0],
    )
    rprims = utils.unpack_primitives(args.rsrf)
    modifiers = {prim.modifier for prim in rprims if prim.ptype in ("polygon", "ring")}
    receiver = Receiver(receiver="", basis=args.basis[1], modifier="")
    sys_paths = args.sys
    for mod in modifiers:
        _receiver = [
            prim
            for prim in rprims
            if prim.modifier == mod and prim.ptype in ("polygon", "ring")
        ]
        if _receiver != []:
            outpath = Path(f"{args.ssrf.stem}_{args.rsrf.stem}.mtx")
            receiver += matrix.surface_as_receiver(
                prim_list=_receiver,
                basis=args.basis[1],
                offset=args.offset[1],
                left=None,
                source="glow",
                out=outpath,
            )
    del args.ssrf, args.rsrf, args.sys, args.basis, args.offset, args.verbose
    matrix.rfluxmtx(sender, receiver, sys_paths, utils.opt2list(vars(args)))


def genmtx_pts_sun(args) -> None:
    """Generate a point to sun matrix."""
    with open(args.pts, "r", encoding="ascii") as rdr:
        sender = matrix.points_as_sender(
            pts_list=[line.split() for line in rdr.readlines()], ray_cnt=1
        )
    full_modifier = False
    receiver = matrix.sun_as_receiver(
        basis=args.basis,
        smx_path=None,
        window_normals=None,
        full_mod=full_modifier,
    )
    outpath = Path(f"{args.pts.stem}_sun.mtx")
    sun_oct = f"sun_{utils.id_generator()}"
    sys_paths = args.sys
    matrix.rcvr_oct(receiver, sys_paths, sun_oct)
    del args.pts, args.basis, args.sys, args.verbose
    matrix.rcontrib(
        sender=sender,
        modifier=receiver.modifier,
        octree=sun_oct,
        out=outpath,
        opt=utils.opt2list(vars(args)),
    )
    os.remove(sun_oct)


def genmtx_vu_sun(args) -> None:
    """Generate a view to sun matrix."""
    with open(args.vu, "r", encoding="ascii") as rdr:
        view = parsers.parse_vu(rdr.readlines()[-1])
        sender = matrix.view_as_sender(
            view,
            ray_cnt=1,
            xres=args.resolu[0],
            yres=args.resolu[1],
        )
    wnormals = None
    if args.window is not None:
        wnormals = utils.primitive_normal(args.window)
    receiver = matrix.sun_as_receiver(
        basis=args.basis,
        smx_path=args.smx_path,
        window_normals=wnormals,
        full_mod=False,
    )
    outpath = Path(f"{args.vu.stem}_sun")
    outpath.mkdir()
    sun_oct = "sun.oct"
    sys_paths = args.sys
    matrix.rcvr_oct(receiver, sys_paths, sun_oct)
    del (
        args.vu,
        args.basis,
        args.sys,
        args.window,
        args.resolu,
        args.smx_path,
        args.verbose,
    )
    with tf.TemporaryDirectory(dir=os.getcwd()) as tempd:
        mod_names = [f"{int(line[3:]):04d}" for line in receiver.modifier.splitlines()]
        matrix.rcontrib(
            sender=sender,
            modifier=receiver.modifier,
            octree=sun_oct,
            out=tempd,
            opt=utils.opt2list(vars(args)),
        )
        _files = sorted(Path(tempd).glob("*.hdr"))
        for idx, val in enumerate(_files):
            val.rename(outpath / (mod_names[idx] + ".hdr"))
        os.remove(sun_oct)


def genmtx_ncp(args: argparse.Namespace) -> None:
    """Generate a matrix/BSDF for a non-coplanar shading systems."""
    sys_paths = args.sys
    wrap = args.wrap
    wprims = utils.unpack_primitives(args.window)
    nprims = utils.unpack_primitives(args.ncp)
    sys_paths.append(args.ncp)
    ports = ncp.gen_port_prims_from_window_ncp(wprims, nprims)
    nmodel = NcpModel(wprims, ports, sys_paths, args.basis[0], args.basis[1])
    out = Path(f"{args.window.stem}_{args.ncp.stem}.mtx")
    del args.window, args.ncp, args.basis, args.sys, args.verbose, args.wrap
    ncp.gen_ncp_mtx(nmodel, out, utils.opt2list(vars(args)), wrap=wrap)


def gen() -> None:
    """Generate things."""
    parser = argparse.ArgumentParser(
        prog="gen",
        description="Generate things.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    prog_subparser = parser.add_subparsers()

    # gen grid
    gengrid_descrp = "Generate an equal-spaced sensor grid based on a polygon."
    parser_grid = prog_subparser.add_parser(
        "grid",
        description=gengrid_descrp,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help=gengrid_descrp,
    )
    parser_grid.set_defaults(func=gengrid)
    parser_grid.add_argument("surface", type=Path, help="surface file path.")
    parser_grid.add_argument("spacing", type=float, help="Grid spacing.")
    parser_grid.add_argument("height", type=float, help="Grid height from surface.")
    parser_grid.add_argument(
        "-op",
        action="store_true",
        help="Generate the grid to the opposite side of surface.",
    )

    # gen glaze
    genglaze_descrp = (
        "Generate a Radiance glazing material based on LBNL IGSDB or Optics file."
    )
    parser_glaze = prog_subparser.add_parser(
        "glaze",
        description=genglaze_descrp,
        help=genglaze_descrp,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_glaze.set_defaults(func=glaze)
    egroup = parser_glaze.add_mutually_exclusive_group(required=True)
    egroup.add_argument("-x", "--optics", nargs="+", type=Path, help="Optics file path")
    egroup.add_argument(
        "-d", "--igsdb", nargs="+", type=Path, help="IGSDB json file path"
    )
    parser_glaze.add_argument(
        "-c", "--cspace", default="radiance", help="Color space for color primaries"
    )
    parser_glaze.add_argument(
        "-s", "--observer", default="2", help="CIE Obvserver 2° or 10°"
    )

    # gen room
    genroom_descrp = "Generate a south-facing side-lit room"
    parser_room = prog_subparser.add_parser(
        "room",
        description=genroom_descrp,
        help=genroom_descrp,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_room.set_defaults(func=genradroom)
    parser_room.add_argument(
        "width", type=float, help="room width along X axis, starting from x=0"
    )
    parser_room.add_argument(
        "depth", type=float, help="room depth along Y axis, starting from y=0"
    )
    parser_room.add_argument(
        "flrflr",
        type=float,
        help="floor to floor height along Z axis, starting from z=0",
    )
    parser_room.add_argument(
        "flrclg",
        type=float,
        help="floor to ceiling height along Z axis, starting from z=0",
    )
    parser_room.add_argument(
        "-w",
        "--window",
        # dest="window",
        metavar=("start_x", "start_z", "width", "height"),
        nargs=4,
        action="append",
        type=float,
        help="Define a window from lower left corner",
    )
    parser_room.add_argument("-n", dest="name", help="Model name", default="model")
    parser_room.add_argument(
        "-t", dest="facade_thickness", metavar="Facade thickness", type=float
    )
    parser_room.add_argument("-r", "--rotate", type=float)

    # gen matrix
    genmatrix_descrp = "Generate various types of matrices."
    parser_matrix = prog_subparser.add_parser(
        "matrix",
        description=genmatrix_descrp,
        help=genmatrix_descrp,
    )
    mtx_subparser = parser_matrix.add_subparsers()

    # gen matrix point-sky
    parser_pts_sky = mtx_subparser.add_parser(
        "point-sky",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_pts_sky.set_defaults(func=genmtx_pts_sky)
    parser_pts_sky.add_argument("pts", type=Path, help="Ray point file path")
    parser_pts_sky.add_argument("sys", nargs="+", type=Path, help="System file path[s]")
    parser_pts_sky.add_argument(
        "-b", "--basis", metavar="", default="r4", help="Sky basis."
    )
    parser_pts_sky = parsers.get_rcontrib_options_args(parser_pts_sky)
    # gen matrix view-sky
    parser_vu_sky = mtx_subparser.add_parser(
        "view-sky",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_vu_sky.set_defaults(func=genmtx_vu_sky)
    parser_vu_sky.add_argument("vu", type=Path, help="View file path.")
    parser_vu_sky.add_argument("sys", nargs="+", type=Path, help="System file path[s].")
    parser_vu_sky.add_argument(
        "-b", "--basis", metavar="", default="r4", help="Sky basis."
    )
    parser_vu_sky.add_argument(
        "-r",
        "--resolu",
        type=int,
        default=(800, 800),
        metavar="",
        help="Image resolution.",
    )
    parser_vu_sky = parsers.get_rcontrib_options_args(parser_vu_sky)
    # gen matrix surface-sky
    parser_srf_sky = mtx_subparser.add_parser(
        "surface-sky",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_srf_sky.set_defaults(func=genmtx_srf_sky)
    parser_srf_sky.add_argument("srf", type=Path, help="Surface file path.")
    parser_srf_sky.add_argument(
        "sys", nargs="+", type=Path, help="System file path[s]."
    )
    parser_srf_sky.add_argument(
        "-b",
        "--basis",
        metavar="",
        nargs=2,
        default=["kf", "r4"],
        help="Surface and sky basis",
    )
    parser_srf_sky.add_argument(
        "-f",
        "--offset",
        metavar="",
        type=float,
        default=0,
        help="Surface offset in its normal direction.",
    )
    parser_srf_sky = parsers.get_rcontrib_options_args(parser_srf_sky)
    # gen matrix point-surface
    parser_pts_srf = mtx_subparser.add_parser(
        "point-surface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_pts_srf.set_defaults(func=genmtx_pts_srf)
    parser_pts_srf.add_argument("pts", type=Path, help="Ray point file path.")
    parser_pts_srf.add_argument("srf", type=Path, help="Surface file path.")
    parser_pts_srf.add_argument(
        "sys", nargs="+", type=Path, help="System file path[s]."
    )
    parser_pts_srf.add_argument(
        "-b", "--basis", default="kf", metavar="", help="Surface basis."
    )
    parser_pts_srf.add_argument(
        "-f",
        "--offset",
        type=float,
        metavar="",
        default=0,
        help="Surface offset in its normal direction.",
    )
    parser_pts_srf = parsers.get_rcontrib_options_args(parser_pts_srf)
    # gen matrix view-surface
    parser_vu_srf = mtx_subparser.add_parser(
        "view-surface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_vu_srf.set_defaults(func=genmtx_vu_srf)
    parser_vu_srf.add_argument("vu", type=Path, help="View file path.")
    parser_vu_srf.add_argument("srf", type=Path, help="Surface file path.")
    parser_vu_srf.add_argument("sys", nargs="+", type=Path, help="System file path[s].")
    parser_vu_srf.add_argument(
        "-b", "--basis", default="kf", metavar="", help="Surface basis."
    )
    parser_vu_srf.add_argument(
        "-r",
        "--resolu",
        nargs=2,
        type=int,
        metavar="",
        default=(800, 800),
        help="Image resolution.",
    )
    parser_vu_srf.add_argument(
        "-f",
        "--offset",
        type=float,
        metavar="",
        default=0,
        help="Surface offset in its normal direction.",
    )
    parser_vu_srf = parsers.get_rcontrib_options_args(parser_vu_srf)
    # gen matrix surface-surface
    parser_srf_srf = mtx_subparser.add_parser(
        "surface-surface",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_srf_srf.set_defaults(func=genmtx_srf_srf)
    parser_srf_srf.add_argument("ssrf", type=Path, help="Sender surface path.")
    parser_srf_srf.add_argument("rsrf", type=Path, help="Receiver surface path.")
    parser_srf_srf.add_argument("sys", nargs="+", type=Path, help="System path[s].")
    parser_srf_srf.add_argument(
        "-b",
        "--basis",
        nargs=2,
        default=["kf", "kf"],
        metavar="",
        help="Sender and receiver basis.",
    )
    parser_srf_srf.add_argument(
        "-f",
        "--offset",
        nargs=2,
        type=float,
        default=[0, 0],
        metavar="",
        help="Sender and receiver offsets in their normal direction.",
    )
    parser_srf_srf = parsers.get_rcontrib_options_args(parser_srf_srf)
    # gen matrix point-sun
    parser_pts_sun = mtx_subparser.add_parser(
        "point-sun",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_pts_sun.set_defaults(func=genmtx_pts_sun)
    parser_pts_sun.add_argument("pts", type=Path, help="Point file path.")
    parser_pts_sun.add_argument(
        "sys", nargs="+", type=Path, help="System file path[s]."
    )
    parser_pts_sun.add_argument(
        "-b", "--basis", default="r6", metavar="", help="Sun basis."
    )
    parser_pts_sun = parsers.get_rcontrib_options_args(parser_pts_sun)
    # gen matrix view-sun
    parser_vu_sun = mtx_subparser.add_parser(
        "view-sun",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_vu_sun.set_defaults(func=genmtx_vu_sun)
    parser_vu_sun.add_argument("vu", type=Path, help="View file path.")
    parser_vu_sun.add_argument("sys", nargs="+", type=Path, help="System file path[s].")
    parser_vu_sun.add_argument(
        "-b", "--basis", default="r6", metavar="", help="Sun basis."
    )
    parser_vu_sun.add_argument(
        "-w", "--window", nargs="+", type=Path, metavar="", help="Window file path[s]."
    )
    parser_vu_sun.add_argument(
        "-s", "--smx_path", type=Path, metavar="", help="Sky matrix file path."
    )
    parser_vu_sun.add_argument(
        "-r",
        "--resolu",
        nargs=2,
        type=int,
        default=(800, 800),
        metavar="",
        help="Image resolution.",
    )
    parser_vu_sun = parsers.get_rcontrib_options_args(parser_vu_sun)
    # gen matrix ncp
    parser_ncp = mtx_subparser.add_parser(
        "ncp",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser_ncp.set_defaults(func=genmtx_ncp)
    parser_ncp.add_argument("window", type=Path, help="Window file paths.")
    parser_ncp.add_argument("ncp", type=Path, help="Non-coplanar shading file paths.")
    parser_ncp.add_argument("sys", nargs="+", type=Path, help="System file path[s].")
    parser_ncp.add_argument(
        "-b",
        "--basis",
        nargs=2,
        default=["kf", "kf"],
        metavar="",
        help="Window and receiving basis.",
    )
    parser_ncp.add_argument(
        "-w",
        "--wrap",
        action="store_true",
        default=False,
        help="Generating a final xml file?",
    )
    parser_ncp = parsers.get_rcontrib_options_args(parser_ncp)

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help=(
            "Verbose mode: -v=Debug -vv=Info -vvv=Warning "
            "-vvvv=Error -vvvvv=Critical"
        ),
    )
    args = parser.parse_args()
    # Setup logger
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s", "%y-%m-%d %H:%M:%S"
    )
    console_handler = logging.StreamHandler()
    _level = args.verbose * 10
    logger.setLevel(_level)
    console_handler.setLevel(_level)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)
    if args.verbose > 1:
        sys.tracebacklimit = 0
    if len(sys.argv) == 1:
        parser.print_usage(sys.stderr)
        sys.exit(1)
    elif len(sys.argv) == 2 and (sys.argv[1] == "matrix"):
        parser_matrix.print_usage(sys.stderr)
        sys.exit(1)
    args.func(args)


def rpxop() -> None:
    """Operate on input directories given a operation type."""
    parser = argparse.ArgumentParser(prog="rpxop", description="Batch image processing")
    subparser = parser.add_subparsers()
    parser_dcts = subparser.add_parser("dctimestep")
    parser_dcts.set_defaults(func=mtxmult.batch_dctimestep)
    parser_dcts.add_argument("mtx", nargs="+", type=Path, help="input matrices")
    parser_dcts.add_argument("sky", type=Path, help="sky files directory")
    parser_dcts.add_argument("out", type=Path, help="output directory")
    parser_dcts.add_argument("-n", type=int, help="number of processors to use")
    parser_pcomb = subparser.add_parser("pcomb")
    parser_pcomb.set_defaults(func=mtxmult.batch_pcomb)
    parser_pcomb.add_argument(
        "inp", type=str, nargs="+", help="list of inputs, e.g., inp1 + inp2.hdr"
    )
    parser_pcomb.add_argument("out", type=Path, help="output directory")
    parser_pcomb.add_argument("-n", type=int, help="number of processors to use")
    args = parser.parse_args()
    if args.func == mtxmult.batch_pcomb:
        inp = [Path(i) for i in args.inp[::2]]
        for i in inp:
            if not i.exists():
                raise FileNotFoundError(i)
        ops = args.inp[1::2]
        args.func(inp, ops, args.out, nproc=args.n)
    elif args.func == mtxmult.batch_dctimestep:
        for i in args.mtx:
            if not i.exists():
                raise FileNotFoundError(i)
        args.func(args.mtx, args.sky, args.out, nproc=args.n)


def genradroom(args) -> None:
    """Commandline interface for generating a generic room.
    Resulting Radiance .rad files will be written to a local
    Objects directory, which will be created if not existed before.
    """
    aroom = room.make_room(
        args.width,
        args.depth,
        args.flrflr,
        args.flrclg,
        args.window,
        swall_thickness=args.facade_thickness,
    )
    name = args.name
    objdir = Path("Objects")
    objdir.mkdir(exist_ok=True)
    with open(objdir / f"materials_{name}.mat", "w", encoding="ascii") as wtr:
        for prim in aroom.materials.values():
            wtr.write(str(prim) + "\n")
    with open(objdir / f"ceiling_{name}.rad", "w", encoding="ascii") as wtr:
        for prim in aroom.ceiling.primitives:
            wtr.write(str(prim) + "\n")
    with open(objdir / f"floor_{name}.rad", "w", encoding="ascii") as wtr:
        for prim in aroom.floor.primitives:
            wtr.write(str(prim) + "\n")
    with open(objdir / f"wall_{name}.rad", "w", encoding="ascii") as wtr:
        for prim in [
            *aroom.swall.primitives,
            *aroom.ewall.primitives,
            *aroom.nwall.primitives,
            *aroom.wwall.primitives,
        ]:
            wtr.write(str(prim) + "\n")
    for idx, srf in enumerate(aroom.swall.windows):
        with open(
            objdir / f"window_{idx:02d}_{name}.rad", "w", encoding="ascii"
        ) as wtr:
            for prim in srf.primitives:
                wtr.write(str(prim) + "\n")



import math
import os
from pathlib import Path
import sys
sys.path.append(".")

from frads import geom
from frads import utils
from frads.types import PaneProperty
from frads.types import PaneRGB
from frads.types import Primitive


test_dir_path = Path(os.path.dirname(__file__))
resource_dir = test_dir_path / "Resources"
check_decimal_to = 6
reinsrc6_path = resource_dir / "reinsrc6.rad"
grid_path = resource_dir / "grid.pts"
prim_path = resource_dir / "model" / "Objects" / "floor_openroom.rad"


pane_property_1 = PaneProperty(
    "VE1-2M_Low-E",  # name
    0.01,  # thickness
    "coated",  # gtype,
    "front",  # coated_side,
    [],  # wavelengths,
    [],  # transmittance,
    [],  # refl_front,
    [],  # refl_back,
)

pane_property_2 = PaneProperty(
    "PVB_laminated",  # name
    0.01,  # thickness
    "laminate",  # gtype,
    "front",  # coated_side,
    [],  # wavelengths,
    [],  # transmittance,
    [],  # refl_front,
    [],  # refl_back,
)

pane_rgb_1 = PaneRGB(
    pane_property_1,
    (0.042, 0.049, 0.043),
    (0.065, 0.058, 0.067),
    (0.756, 0.808, 0.744),
)

pane_rgb_2 = PaneRGB(
    pane_property_2,
    (0.11, 0.11, 0.11),
    (0.11, 0.11, 0.11),
    (0.63, 0.63, 0.63),
)

def test_polygon2prim():
    pass

def test_primitive_normal():
    pass

def test_samp_dir():
    pass

def test_up_vector():
    pass

def test_neutral_plastic_prim():
    pass

def test_neutral_trans_prim():
    pass

def test_color_plastic_prim():
    pass

def test_glass_prim():
    pass

def test_bsdf_prim():
    pass

def test_lambda_calc():
    pass

def test_angle_basis_coeff():
    pass

def test_opt2list():
    opt1 = {"ab": 2, "ad": 1024, "I": True}
    res1 = utils.opt2list(opt1)
    answer1 = ["-ab", "2", "-ad", "1024", "-I+"]
    assert res1 == answer1
    opt2 = {"vt": "a", "vf": "test/v1a.vf", "vp":[0, 0, 0]}
    res2 = utils.opt2list(opt2)
    answer2 = ["-vta", "-vf", "test/v1a.vf", "-vp", "0", "0", "0"]
    assert res2 == answer2
    opt3 = {"ab": 3, "ad": 64, "av": [.38, .38, .38], "u": True, "vt": "a"}
    res3 = utils.opt2list(opt3)
    answer3 = ["-ab", "3", "-ad", "64",
               "-av", "0.38", "0.38", "0.38", "-u+", "-vta"]
    assert res3 == answer3


def test_calc_reinsrc_dir():
    mf = 6
    vecs, omgs = utils.calc_reinsrc_dir(mf)
    with open(reinsrc6_path) as rdr:
        answer_lines = rdr.readlines()
    answer_vecs = []
    answer_omgs = []
    for vec, omg, line in zip(vecs, omgs, answer_lines):
        x, y, z, aomg = map(float, line.strip().split())
        assert round(vec.x, 3) == round(x, 3)
        assert round(vec.y, 3) == round(y, 3)
        assert round(vec.z, 3) == round(z, 3)
        assert round(omg, 3) == round(aomg, 3)

def test_pt_inclusion():
    pass

def test_gen_grid():
    polygon = geom.Polygon([
        geom.Vector(0, 0, 0),
        geom.Vector(0, 14, 0),
        geom.Vector(12, 14, 0),
        geom.Vector(12, 0, 0),
    ])
    height = 0.76
    spacing = 1
    result = utils.gen_grid(polygon, height, spacing)
    with open(grid_path) as rdr:
        lines = rdr.readlines()
    assert len(result) == len(lines)
    for res, ans in zip(result, lines):
        for r, a in zip(res, ans.split()):
            assert r == float(a)


def test_gen_blinds():
    pass

def test_analyze_vert_polygon():
    pass

def test_varays():
    pass

def test_get_glazing_primitive():
    prim = utils.get_glazing_primitive([pane_rgb_1, pane_rgb_2])
    res = str(prim).strip().replace("\n", " ")
    answer = ("void BRTDfunc VE1-2M_Low-E+PVB_laminated 10 if(Rdot,cr(fr("
              "0.110),ft(0.630),fr(0.065)),cr(fr(0.042),ft(0.756),fr(0.110)))"
              " if(Rdot,cr(fr(0.110),ft(0.630),fr(0.058)),cr(fr(0.049),ft("
              "0.808),fr(0.110))) if(Rdot,cr(fr(0.110),ft(0.630),fr(0.067)),"
              "cr(fr(0.043),ft(0.744),fr(0.110))) ft(0.630)*ft(0.756) ft("
              "0.630)*ft(0.808) ft(0.630)*ft(0.744) 0 0 0 glaze2.cal 0 9 0 "
              "0 0 0 0 0 0 0 0")
    assert res == answer
    prim2 = utils.get_glazing_primitive([pane_rgb_1])
    res2 = str(prim2).strip().replace("\n", " ")
    answer2 = ("void BRTDfunc VE1-2M_Low-E "
               "10 sr_clear_r sr_clear_g sr_clear_b "
               "st_clear_r st_clear_g st_clear_b 0 0 0 "
               "glaze1.cal 0 19 0 0 0 0 0 0 0 0 0 "
               "1 0.065 0.058 0.067 0.042 0.049 0.043 "
               "0.756 0.808 0.744")
    assert res2 == answer2

def test_unpack_primitives():
    prims = utils.unpack_primitives(prim_path)
    assert isinstance(prims[0], Primitive)

def test_nest_list():
    pass

def test_write_square_matrix():
    pass

def test_dhi2dni():
    pass

def test_is_number():
    pass

def test_silent_remove():
    pass

def test_square2disk():
    pass

def test_id_generator():
    pass

def get_nested_list_levels():
    pass



from pathlib import Path
import sys
sys.path.append(".")

from frads import geom
from frads import matrix
from frads.types import Primitive
from frads.types import View
from frads.types import ViewType

window_polygon = [
    geom.Polygon([geom.Vector(0, 0, 0),
                  geom.Vector(0, 0, 3),
                  geom.Vector(2, 0, 3),
                  geom.Vector(2, 0, 0),
                  ]),
    geom.Polygon([geom.Vector(3, 0, 0),
                  geom.Vector(3, 0, 3),
                  geom.Vector(5, 0, 3),
                  geom.Vector(5, 0, 0),
                  ])
]
window_primitives = [
    Primitive("void", "polygon", "window1", ("0"), window_polygon[0].to_real()),
    Primitive("void", "polygon", "window2", ("0"), window_polygon[1].to_real())
]

def test_surface_as_sender():
    basis = "kf"
    sender = matrix.surface_as_sender(
        window_primitives, basis, offset=None, left=None)
    assert sender.form == "s"
    assert sender.xres is None
    assert sender.yres is None

def test_view_as_sender():
    view = View(
        position=geom.Vector(0,  0, 0),
        direction=geom.Vector(0, -1, 0),
        hori=180,
        vert=180,
        vtype=ViewType.angular_fisheye,
        xres=4,
        yres=4,
    )
    ray_cnt = 5
    sender = matrix.view_as_sender(view, ray_cnt, view.xres, view.yres)
    assert sender.form == "v"
    assert sender.xres == view.xres
    assert sender.yres == view.yres

def test_point_as_sender():
    pts_list = [[0,0,0,0,0,1], [0,0,3,0,0,1]]
    ray_cnt = 5
    sender = matrix.points_as_sender(pts_list, ray_cnt)
    assert sender.form == "p"
    assert sender.xres is None
    assert sender.yres == len(pts_list) * ray_cnt

def test_surface_as_receiver():
    basis = "kf"
    out = None
    offset = 0.1
    receiver = matrix.surface_as_receiver(
        window_primitives, basis, out, offset=offset)
    assert receiver.receiver != ""
    assert receiver.basis == 'kf'
    assert receiver.modifier == ""


def test_sky_as_receiver():
    basis = 'r1'
    out = Path("test.mtx")
    receiver = matrix.sky_as_receiver(basis, out)
    assert receiver.receiver != ""
    assert receiver.basis == 'r1'
    assert receiver.modifier == ''


def test_sun_as_receiver():
    basis = 'r6'
    smx_path = None
    window_normals = None
    receiver = matrix.sun_as_receiver(basis, smx_path, window_normals)
    assert receiver.receiver != ""
    assert receiver.basis == "r6"
    assert receiver.modifier != ""





"""Generic room model"""
from typing import List
from typing import Optional

from frads import geom
from frads import utils
from frads.types import Primitive


class Surface:
    """Surface object."""

    def __init__(self, base: geom.Polygon) -> None:
        """."""
        self.base = base
        self._vertices = base.vertices
        self._vect1 = (base.vertices[1] - base.vertices[0]).normalize()
        self._vect2 = (base.vertices[2] - base.vertices[1]).normalize()
        self.polygons: List[geom.Polygon] = [self.base]
        self.windows: List[Surface] = []
        self._modifier: str = "void"
        self._identifier: str = "void"
        self._primitives: Optional[List[Primitive]] = None

    @property
    def modifier(self):
        """."""
        return self._modifier

    @property
    def identifier(self):
        """."""
        return self._identifier

    @modifier.setter
    def modifier(self, mod):
        """."""
        self._modifier = mod

    @identifier.setter
    def identifier(self, identifier):
        """."""
        self._identifier = identifier

    @property
    def primitives(self) -> Optional[List[Primitive]]:
        """."""
        self._primitives = []
        for idx, polygon in enumerate(self.polygons):
            self._primitives.append(
                Primitive(
                    self.modifier,
                    "polygon",
                    f"{self.identifier}_{idx:02d}",
                    ["0"],
                    polygon.to_real(),
                )
            )
        return self._primitives

    def make_window_wwr(self, wwr: float) -> None:
        """Make a window based on window-to-wall ratio."""
        window_polygon = self.base.scale(geom.Vector(*[wwr] * 3), self.base.centroid)
        self.base = self.base - window_polygon
        self.windows.append(Surface(window_polygon))

    def make_window(
        self, dist_left: float, dist_bot: float, width: float, height: float
    ) -> None:
        """Make a window and punch a hole."""
        win_pt1 = (
            self._vertices[0]
            + self._vect1.scale(dist_bot)
            + self._vect2.scale(dist_left)
        )
        win_pt2 = win_pt1 + self._vect1.scale(height)
        win_pt3 = win_pt1 + self._vect2.scale(width)
        window_polygon = geom.Polygon.rectangle3pts(win_pt3, win_pt1, win_pt2)
        self.base = self.base - window_polygon
        self.windows.append(Surface(window_polygon))

    def thicken(self, thickness: float) -> None:
        """Thicken the surface."""
        direction = self.base.normal.scale(thickness)
        polygons = self.base.extrude(direction)
        counts = [polygons.count(plg) for plg in polygons]
        self.polygons = [plg for plg, cnt in zip(polygons, counts) if cnt == 1]

    def move_window(self, distance: float) -> None:
        """Move windows in its normal direction."""
        direction = self.base.normal.scale(distance)
        self.windows = [Surface(window.base.move(direction)) for window in self.windows]

    def rotate(self, deg):
        """Rotate the surface counter clock-wise."""
        polygons = []
        for plg in self.polygons:
            polygons.append(plg.rotate3d(deg, geom.Vector(0, 0, 1)))
        self.polygons = polygons
        for window in self.windows:
            wpolygons = []
            for plg in window.polygons:
                wpolygons.append(plg.rotate3d(deg, geom.Vector(0, 0, 1)))


class Room:
    """Make a shoebox."""

    def __init__(self, floor: Surface, ceiling, swall, ewall, nwall, wwall) -> None:
        """."""
        self.floor = floor
        self.ceiling = ceiling
        self.swall = swall
        self.ewall = ewall
        self.nwall = nwall
        self.wwall = wwall
        self.materials = utils.material_lib()

    @classmethod
    def from_wdh(
        cls,
        width: float,
        depth: float,
        floor_floor: float,
        floor_ceiling: float,
        origin: Optional[geom.Vector] = None,
    ) -> "Room":
        """Generate a room from width, depth, and height."""
        pt1 = geom.Vector(0, 0, 0) if origin is None else origin
        pt2 = pt1 + geom.Vector(width, 0, 0)
        pt3 = pt2 + geom.Vector(0, depth, 0)
        floor = geom.Polygon.rectangle3pts(pt1, pt2, pt3)
        _, ceiling, swall, ewall, nwall, wwall = floor.extrude(
            geom.Vector(0, 0, floor_floor)
        )
        ceiling = ceiling.move(geom.Vector(0, 0, floor_ceiling - floor_floor))
        return cls(
            Surface(floor),
            Surface(ceiling),
            Surface(swall),
            Surface(ewall),
            Surface(nwall),
            Surface(wwall),
        )

    @property
    def primitives(self):
        """."""
        return [
            *self.materials.values(),
            *self.floor.primitives,
            *self.ceiling.primitives,
            *self.swall.primitives,
            *self.ewall.primitives,
            *self.nwall.primitives,
            *self.wwall.primitives,
        ]

    @property
    def window_primitives(self):
        """."""
        return [
            *[prim for srf in self.ceiling.windows for prim in srf.primitives],
            *[prim for srf in self.swall.windows for prim in srf.primitives],
            *[prim for srf in self.ewall.windows for prim in srf.primitives],
            *[prim for srf in self.nwall.windows for prim in srf.primitives],
            *[prim for srf in self.wwall.windows for prim in srf.primitives],
        ]

    def get_material_names(self) -> List[str]:
        """Get material identifiers."""
        return [prim.identifier for prim in self.materials.values()]

    def add_material(self, primitive) -> None:
        """Add a material to the material library."""
        self.materials[primitive.identifier] = primitive

    def validate(self) -> None:
        """Validate the room model."""
        for prim in [
            *self.floor.primitives,
            *self.ceiling.primitives,
            *self.swall.primitives,
            *self.ewall.primitives,
            *self.nwall.primitives,
            *self.wwall.primitives,
        ]:
            if prim.modifier not in self.materials:
                raise ValueError(
                    f"Unknown modifier {prim.modifier} in {prim.identifier}"
                )

    def rotate(self, deg):
        """Rotate the room counter clock-wise."""
        self.floor.rotate(deg)
        self.ceiling.rotate(deg)
        self.swall.rotate(deg)
        self.ewall.rotate(deg)
        self.wwall.rotate(deg)
        self.nwall.rotate(deg)


def make_room(
    width: float,
    depth: float,
    floor_floor: float,
    floor_ceiling: float,
    windows,
    swall_thickness=None,
):
    """Make a side-lit shoebox room as a Room object."""
    aroom = Room.from_wdh(width, depth, floor_floor, floor_ceiling)
    if windows is not None:
        for window in windows:
            aroom.swall.make_window(*window)
        for window in aroom.swall.windows:
            window.modifier = "glass_60"
    if swall_thickness is not None:
        aroom.swall.thicken(swall_thickness)
    aroom.swall.modifier = "neutral_lambertian_0.5"
    aroom.ewall.modifier = "neutral_lambertian_0.5"
    aroom.nwall.modifier = "neutral_lambertian_0.5"
    aroom.wwall.modifier = "neutral_lambertian_0.5"
    aroom.ceiling.modifier = "neutral_lambertian_0.7"
    aroom.floor.modifier = "neutral_lambertian_0.2"
    return aroom



"""
This module contains all data parsing routines.
"""
import argparse
import configparser
from datetime import datetime
from pathlib import Path
import re
import subprocess as sp
from typing import Tuple, Any
from typing import Dict
from typing import Generator
from typing import List
from typing import Sequence
from typing import Union

from frads import geom
from frads.types import Primitive
from frads.types import PaneProperty
from frads.types import View
from frads.types import WeaMetaData
from frads.types import WeaData


def parse_mrad_config(cfg_path: Path) -> configparser.ConfigParser:
    """Parse mrad configuration file."""
    if not cfg_path.is_file():
        raise FileNotFoundError(cfg_path)
    config = configparser.ConfigParser(
        allow_no_value=False,
        inline_comment_prefixes="#",
        interpolation=configparser.ExtendedInterpolation(),
        converters={
            "path": lambda x: Path(x.strip()),
            "paths": lambda x: [Path(i) for i in x.split()],
            "options": parse_opt,
            "view": parse_vu,
        },
    )
    config.read(Path(__file__).parent / "data" / "mrad_default.cfg")
    config.read(cfg_path)
    return config


def parse_epw(epw_str: str) -> tuple:
    """Parse epw file and return wea header and data."""
    raw = epw_str.splitlines()
    epw_header = raw[0].split(",")
    content = raw[8:]
    data = []
    for li in content:
        line = li.split(",")
        year = int(line[0])
        month = int(line[1])
        day = int(line[2])
        hour = int(line[3]) - 1
        hours = hour + 0.5
        dir_norm = float(line[14])
        dif_hor = float(line[15])
        cc = float(line[19])
        aod = float(line[26])
        data.append(
            WeaData(datetime(year, month, day, hour, 30), dir_norm, dif_hor, cc, aod)
        )
    city = epw_header[1]
    country = epw_header[3]
    latitude = float(epw_header[6])
    longitude = -1 * float(epw_header[7])
    tz = int(float(epw_header[8])) * (-15)
    elevation = float(epw_header[9].rstrip())
    meta_data = WeaMetaData(city, country, latitude, longitude, tz, elevation)
    return meta_data, data


def parse_idf(content: str) -> dict:
    """Parse an IDF file into a dictionary."""
    sections = content.rstrip().split(";")
    sub_sections: List[List[str]] = []
    obj_dict: Dict[str, List[List[str]]] = {}
    for sec in sections:
        sec_lines = sec.splitlines()
        _lines = []
        for sl in sec_lines:
            content = sl.split("!")[0]
            if content != "":
                _lines.append(content)
        _lines = " ".join(_lines).split(",")
        clean_lines = [i.strip() for i in _lines]
        sub_sections.append(clean_lines)

    for ssec in sub_sections:
        obj_dict[ssec[0].lower()] = []
    for ssec in sub_sections:
        obj_dict[ssec[0].lower()].append(ssec[1:])
    return obj_dict


def parse_igsdb_json(json_obj: dict) -> PaneProperty:
    """Parse a JSON file from IGSDB."""
    name = json_obj["name"].replace(" ", "_")
    gtype = json_obj["type"]
    coated_side = json_obj["coated_side"].lower()
    thickness = json_obj["measured_data"]["thickness"]
    spectral_data = json_obj["spectral_data"]["spectral_data"]

    wavelength = []
    transmittance = []
    reflectance_front = []
    reflectance_back = []

    for row in spectral_data:
        wavelength.append(row["wavelength"] * 1e3)  # um to nm
        transmittance.append(row["T"])
        reflectance_front.append(row["Rf"])
        reflectance_back.append(row["Rb"])
    return PaneProperty(
        name,
        thickness,
        gtype,
        coated_side,
        wavelength,
        transmittance,
        reflectance_front,
        reflectance_back,
    )


def get_rcontrib_options_args(parser):
    """Add rcontrib specific options to a parser."""
    parser.add_argument(
        "-I", action="store_true", dest="I", default=None, help="Toggle Ill"
    )
    parser.add_argument(
        "-I+", action="store_true", dest="I", default=None, help="Ill. On"
    )
    parser.add_argument(
        "-I-", action="store_false", dest="I", default=None, help="Ill. Off"
    )
    parser.add_argument(
        "-i", action="store_true", dest="i", default=None, help="Toggle ill(final)"
    )
    parser.add_argument(
        "-i+", action="store_true", dest="i", default=None, help="Ill(final). On"
    )
    parser.add_argument(
        "-i-", action="store_false", dest="i", default=None, help="Ill(final). Off"
    )
    parser.add_argument(
        "-V",
        action="store_true",
        dest="V",
        default=None,
        help="Toggle to compute cofficent",
    )
    parser.add_argument(
        "-V+",
        action="store_true",
        dest="V",
        default=None,
        help="Use actual source value",
    )
    parser.add_argument(
        "-V-", action="store_false", dest="V", default=None, help="Compute coefficients"
    )
    parser.add_argument("-ab", type=int, metavar="", help="ambient bounces")
    parser.add_argument("-ad", type=int, metavar="", help="Ambient division")
    parser.add_argument("-c", type=int, metavar="", help="ray count")
    parser.add_argument("-dc", type=int, metavar="", help="direct certainty")
    parser.add_argument("-dj", type=float, metavar="", help="direct jitter")
    parser.add_argument("-dp", type=int, metavar="", help="direct pixel")
    parser.add_argument("-dr", type=int, metavar="", help="direct rely")
    parser.add_argument("-ds", type=int, metavar="", help="direct sampling")
    parser.add_argument("-dt", type=int, metavar="", help="direct smapling threshol")
    parser.add_argument("-lr", type=int, metavar="", help="reflectance limits")
    parser.add_argument("-lw", type=float, metavar="", help="limit weight")
    parser.add_argument("-ss", type=int, metavar="", help="specular sampling")
    parser.add_argument("-st", type=int, metavar="", help="specular sampling threshold")
    return parser


def get_rtrace_options_args(parser):
    """Add rtrace options and flags to a parser."""
    parser.add_argument("-u", action="store_true", dest="u", default=None)
    parser.add_argument("-u+", action="store_true", dest="u", default=None)
    parser.add_argument("-u-", action="store_false", dest="u", default=None)
    parser.add_argument("-ld", dest="ld", help="Limit ray distance")
    parser.add_argument(
        "-ld-",
        action="store_false",
        dest="ld",
        default=None,
        help="Unlimit ray distance",
    )
    parser.add_argument("-aa", type=float, metavar="", help="ambient accuracy")
    parser.add_argument("-ar", type=int, metavar="", help="ambient resolution")
    parser.add_argument("-as", type=int, metavar="", help="ambient super sampling")
    parser.add_argument("-av", type=float, nargs=3, metavar="", help="ambient values")
    parser = get_rcontrib_options_args(parser)
    return parser


def parse_opt(opt_str: str) -> dict:
    """Parsing option string into a dictionary.

    Args:
        opt_str: rtrace option parameters as a string

    Returns:
        An option dictionary
    """

    args_list = opt_str.strip().split()
    oparser = argparse.ArgumentParser()
    oparser.add_argument("-w", action="store_true", dest="w", default=None)
    oparser.add_argument("-w+", action="store_true", dest="w", default=None)
    oparser.add_argument("-w-", action="store_false", dest="w", default=None)
    oparser.add_argument("-f", action="store")
    oparser.add_argument("-hd", action="store_const", const="", default=None)
    oparser.add_argument("-n", type=int)
    oparser = get_rtrace_options_args(oparser)
    args, _ = oparser.parse_known_args(args_list)
    opt_dict = vars(args)
    opt_dict = {k: v for (k, v) in opt_dict.items() if v is not None}
    return opt_dict


def parse_optics(fpath) -> PaneProperty:
    """Read and parse an optics file."""
    # enc = 'cp1250' #decoding needed to parse header
    with open(fpath, errors="ignore") as rdr:
        raw = rdr.read()
    header_lines = [i for i in raw.splitlines() if i.startswith("{")]
    if header_lines == []:
        raise Exception("No header in optics file")
    header = {}
    for line in header_lines:
        if line.strip().split("}")[-1] != "":
            key = re.search("{(.*?)}", line).group(1).strip()
            val = line.split("}")[-1].strip()
            header[key] = val
        elif line:
            content = re.search("{(.*?)}", line).group(1).strip()
            if content != "":
                key = content.split(":")[0].strip()
                val = content.split(":")[1].strip()
                header[key] = val
    name = header["Product Name"].replace(" ", "_")
    thickness = float(header["Thickness"])
    gtype = header["Type"]
    coated_side = header["Coated Side"].lower()
    data = [i.split() for i in raw.strip().splitlines() if not i.startswith("{")]
    wavelength = [float(row[0]) for row in data]
    transmittance = [float(row[1]) for row in data]
    reflectance_front = [float(row[2]) for row in data]
    reflectance_back = [float(row[3]) for row in data]
    if header["Units, Wavelength Units"] == "SI Microns":  # um to nm
        wavelength = [val * 1e3 for val in wavelength]
    return PaneProperty(
        name,
        thickness,
        gtype,
        coated_side,
        wavelength,
        transmittance,
        reflectance_front,
        reflectance_back,
    )


def parse_polygon(real_args: Sequence[Union[int, float]]) -> geom.Polygon:
    """Parse real arguments to polygon.
    Args:
        primitive: a dictionary object containing a primitive

    Returns:
        modified primitive
    """
    coords = real_args[1:]
    arg_cnt = int(real_args[0])
    vertices = [geom.Vector(*coords[i : i + 3]) for i in range(0, arg_cnt, 3)]
    return geom.Polygon(vertices)


def parse_primitive(lines: list) -> List[Primitive]:
    """Parse Radiance primitives inside a file path into a list of dictionary.
    Args:
        lines: list of lines as strings

    Returns:
        list of primitives as dictionaries

    Notes:
        Dropping support for alias type
    """
    # Expand in-line commands
    cmd_lines = [(idx, line) for idx, line in enumerate(lines) if line.startswith("!")]
    cmd_results = []
    for cmd in cmd_lines:
        cmd_results.append(
            sp.run(cmd[1][1:], check=True, stdout=sp.PIPE).stdout.decode().splitlines()
        )
    counter = 0
    for idx, item in enumerate(cmd_lines):
        counter += item[0]
        lines[counter : counter + 1] = cmd_results[idx]
        counter += len(cmd_results[idx]) - 1 - item[0]

    content = " ".join(
        [i.strip() for i in lines if i.strip() != "" and i[0] != "#"]
    ).split()
    primitives: List[Primitive] = []
    idx = 0
    while idx < len(content):
        _modifier = content[idx]
        _type = content[idx + 1]
        if _type == "alias":
            # _name_to = content[idx + 2]
            # _name_from = content[idx + 3]
            # primitives.append(
            # Alias(_modifier, _name_to, _name_from)
            # )
            idx += 4
            continue
        _identifier = content[idx + 2]
        str_arg_cnt = int(content[idx + 3])
        _str_args = content[idx + 3 : idx + 4 + str_arg_cnt]
        idx += 5 + str_arg_cnt
        real_arg_cnt = int(content[idx])
        _real_args = [float(i) for i in content[idx : idx + 1 + real_arg_cnt]]
        idx += real_arg_cnt + 1
        primitives.append(
            Primitive(_modifier, _type, _identifier, _str_args, _real_args)
        )
    return primitives


def parse_rad_header(header_str: str) -> tuple:
    """Parse a Radiance matrix file header.

    Args:
        header_str(str): header as string
    Returns:
        A tuple contain nrow, ncol, ncomp, datatype
    Raises:
        ValueError if any of NROWs NCOLS NCOMP FORMAT is not found.
        (This is problematic as it can happen)
    """
    compiled = re.compile(
        r" NROWS=(.*) | NCOLS=(.*) | NCOMP=(.*) | FORMAT=(.*) ", flags=re.X
    )
    matches = compiled.findall(header_str)
    if len(matches) != 4:
        raise ValueError("Can't find one of the header entries.")
    nrow = int([mat[0] for mat in matches if mat[0] != ""][0])
    ncol = int([mat[1] for mat in matches if mat[1] != ""][0])
    ncomp = int([mat[2] for mat in matches if mat[2] != ""][0])
    dtype = [mat[3] for mat in matches if mat[3] != ""][0].strip()
    return nrow, ncol, ncomp, dtype


def parse_vu(vu_str: str) -> View:
    """Parse view string into a View object.

    Args:
        vu_str: view parameters as a string

    Returns:
        A view object
    """

    if vu_str.strip() == "":
        return
    args_list = vu_str.strip().split()
    vparser = argparse.ArgumentParser()
    vparser.add_argument("-v", action="store", dest="vt")
    vparser.add_argument("-vp", nargs=3, type=float)
    vparser.add_argument("-vd", nargs=3, type=float)
    vparser.add_argument("-vu", nargs=3, type=float)
    vparser.add_argument("-vv", type=float)
    vparser.add_argument("-vh", type=float)
    vparser.add_argument("-vo", type=float)
    vparser.add_argument("-va", type=float)
    vparser.add_argument("-vs", type=float)
    vparser.add_argument("-vl", type=float)
    vparser.add_argument("-x", type=int)
    vparser.add_argument("-y", type=int)
    vparser.add_argument("-vf", type=argparse.FileType("r"))
    args, _ = vparser.parse_known_args(args_list)
    if args.vf is not None:
        args, _ = vparser.parse_known_args(
            args.vf.readline().strip().split(), namespace=args
        )
        args.vf.close()
    if None in (args.vp, args.vd):
        raise ValueError("Invalid view")
    view = View(geom.Vector(*args.vp), geom.Vector(*args.vd))
    if args.vt is not None:
        view.vtype = args.vt[-1]
    if args.x is not None:
        view.xres = args.x
    if args.y is not None:
        view.yres = args.y
    if args.vv is not None:
        view.vert = args.vv
    if args.vh is not None:
        view.hori = args.vh
    if args.vo is not None:
        view.fore = args.vo
    if args.va is not None:
        view.aft = args.va
    if args.vs is not None:
        view.shift = args.vs
    if args.vl is not None:
        view.lift = args.vl
    return view


def tokenize(inp: str) -> Generator[str, None, None]:
    """Generator for tokenizing a string that
    is seperated by a space or a comma.
    Args:
       inp: input string
    Yields:
        next token
    """
    tokens = re.compile(
        " +|[-+]?(\d+([.,]\d*)?|[.,]\d+)([eE][-+]?\d+)+|[\d*\.\d+]+|[{}]"
    )
    for match in tokens.finditer(inp):
        if match.group(0)[0] in " ,":
            continue
        yield match.group(0)


def parse_branch(token: Generator[str, None, None]) -> Any:
    """Prase tensor tree branches recursively by opening and closing curly braces.
    Args:
        token: token generator object.
    Return:
        children: parsed branches as nexted list
    """
    children = []
    while True:
        value = next(token)
        if value == "{":
            children.append(parse_branch(token))
        elif value == "}":
            return children
        else:
            children.append(float(value))


def parse_ttree(data_str: str) -> list:
    """Parse a tensor tree.
    Args:
        data_str: input data string
    Returns:
        A nested list that is the tree
    """
    tokenized = tokenize(data_str)
    if next(tokenized) != "{":
        raise ValueError("Tensor tree data not starting with {")
    return parse_branch(tokenized)


def get_nested_list_levels(nested_list: list) -> int:
    """Calculate the number of levels given a nested list."""
    return (
        isinstance(nested_list, list)
        and max(map(get_nested_list_levels, nested_list)) + 1
    )


class TensorTree:
    """The tensor tree object.
    Anisotropic tensor tree has should have 16 lists
    Attributes:
        parsed: parsed tensor tree object)
        depth: number of tree levels
    """

    def __init__(self, parsed) -> None:
        self.parsed = parsed
        self.depth = get_nested_list_levels(parsed)

    def lookup(self, xp, yp) -> list:
        """Traverses a parsed tensor tree (a nexted list) given a input position."""
        branch_idx = self.get_branch_index(xp, yp)
        quads = [self.parsed[i] for i in branch_idx]
        return [self.traverse(quad, xp, yp) for quad in quads]

    def get_leaf_index(self, xp, yp) -> range:
        if xp < 0:
            if yp < 0:
                return range(0, 4)
            return range(4, 8)
        if yp < 0:
            return range(8, 12)
        return range(12, 16)

    def get_branch_index(self, xp, yp) -> range:
        """Gets a set of index."""
        if xp < 0:
            if yp < 0:
                return range(0, 16, 4)
            return range(2, 16, 4)
        if yp < 0:
            return range(1, 16, 4)
        return range(3, 16, 4)

    def traverse(self, quad, xp, yp, n: int = 1) -> list:
        """Traverse a quadrant."""
        if len(quad) == 1:  # single leaf
            res = quad
        else:
            res = []
            # get x, y signage in relation to branches
            _x = xp + 1 / (2**n) if xp < 0 else xp - 1 / (2**n)
            _y = yp + 1 / (2**n) if yp < 0 else yp - 1 / (2**n)
            n += 1
            # which four branches to get? get index for them
            if n < self.depth:
                ochild = self.get_branch_index(_x, _y)
            else:
                ochild = self.get_leaf_index(_x, _y)
            sub_quad = [quad[i] for i in ochild]
            if all(isinstance(i, float) for i in sub_quad):
                res = sub_quad  # single leaf for each branch
            else:  # branches within this quadrant
                for sq in sub_quad:
                    if len(sq) > 4:  # there is another branch
                        res.append(self.traverse(sq, _x, _y, n=n))
                    else:  # just a leaf
                        res.append(sq)
        return res


def parse_wea(wea_str: str) -> Tuple[WeaMetaData, List[WeaData]]:
    """Parse a wea file in its entirety."""
    lines = wea_str.splitlines()
    place = lines[0].split(" ", 1)[1]
    lat = float(lines[1].split(" ", 1)[1])
    lon = float(lines[2].split(" ", 1)[1])
    tz = int(float(lines[3].split(" ", 1)[1]))
    ele = float(lines[4].split(" ", 1)[1])
    meta_data = WeaMetaData(place, "", lat, lon, tz, ele)
    year = datetime.today().year
    data = []
    for li in lines[6:]:
        if li.strip() == "":
            continue
        line = li.split()
        month = int(line[0])
        day = int(line[1])
        hours = float(line[2])
        hour = int(hours)
        minute = int((hours - hour) * 60)
        dir_norm = float(line[3])
        dif_hor = float(line[4])
        data.append(
            WeaData(datetime(year, month, day, hour, minute), dir_norm, dif_hor)
        )
    return meta_data, data



"""Multiply Radiance matrix files.
"""

import logging
import os
from pathlib import Path
import subprocess as sp
from typing import Tuple, List
from typing import Optional

from frads import parsers
from frads import utils

try:
    import numpy as np

    NUMPY_FOUND = True
except ModuleNotFoundError:
    NUMPY_FOUND = False


logger: logging.Logger = logging.getLogger("frads.mtxmult")


def batch_dctimestep(
    mtx: List[Path], sky_dir: Path, out_dir: Path, nproc: Optional[int] = None
) -> None:
    """Run dctimestep in batches."""
    out_dir.mkdir(parents=True, exist_ok=True)
    nproc = 1 if nproc is None else nproc
    if len(mtx) == 0:
        raise ValueError("Input matrices files empty")
    if not sky_dir.is_dir():
        raise ValueError("Sky directory not exist")
    cmds: List[List[str]] = []
    outs: List[str] = []
    for sky in sorted(sky_dir.glob("*")):
        out_path = out_dir / sky.with_suffix(".hdr")
        cmd = ["dctimestep"] + [str(f) for f in mtx]
        cmd.append(str(sky))
        cmds.append(cmd)
        outs.append(out_path)
    utils.batch_process(cmds, opaths=outs, nproc=nproc)


def batch_pcomb(
    inp: List[Path], ops: List[str], out_dir: Path, nproc: Optional[int] = None
) -> None:
    """
    inputs: e.g.[inpdir1, inpdir2, inpdir3.hdr]
    ops: e.g.['+', '-']
    out_dir: output directory
    Resulting files will be named after first input item.
    """
    out_dir.mkdir(exist_ok=True)
    nproc = 1 if nproc is None else nproc
    if (len(inp) < 2) or (len(ops) < 1):
        raise ValueError("Invalid # of inputs")
    if inp[0].is_file():
        raise ValueError("First input cannot be a file, please reformulate")
    expanded_inp: List[List[Path]] = []
    max_len = 0
    for i in inp:
        if i.is_dir():
            hdrs = sorted(i.glob("*.hdr"))
            max_len = max(max_len, len(hdrs))
            expanded_inp.append(list(hdrs))
        elif i.is_file():
            expanded_inp.append([i])
    if max_len > 1:
        for idx, ip in enumerate(expanded_inp):
            if len(ip) == 1:
                expanded_inp[idx] = ip * max_len
    equal_len = all(len(i) == len(expanded_inp[0]) for i in expanded_inp)
    if not equal_len:
        logger.warning("Input directories don't the same number of files")
    grouped = [list(i) for i in zip(*expanded_inp)]
    rstr_list = [f"ri({i+1})" for i, _ in enumerate(expanded_inp)]
    frstr_list = rstr_list + ops
    frstr_list[::2] = rstr_list
    frstr_list[1::2] = ops
    frstr_list.insert(0, "ro=")
    frstr = "".join(frstr_list)
    fgstr = frstr.replace("r", "g")
    fbstr = frstr.replace("r", "b")
    pcomb_expr = frstr + ";" + fgstr + ";" + fbstr
    cmds = []
    opaths = []
    for group in grouped:
        cmd = ["pcomb", "-e", pcomb_expr]
        for cp in group:
            cmd += ["-o", str(cp)]
        cmds.append(cmd)
        opaths.append(out_dir / group[0].name)
    utils.batch_process(cmds, opaths=opaths, nproc=nproc)


def mtxstr2nparray(data_str: bytes):
    """Convert Radiance 3-channel matrix file to numpy array.

    Args:
        data_str: file data string

    Returns:
        RGB numpy arrays
    """
    if not data_str.startswith(b"#?RADIANCE"):
        raise ValueError("No header found")
    if b"\r\n\r\n" in data_str:
        linesep2 = b"\r\n\r\n"
    else:
        linesep2 = b"\n\n"
    chunks = data_str.split(linesep2, 1)
    nrow, ncol, ncomp, dtype = parsers.parse_rad_header(chunks[0].decode())
    if dtype == "ascii":
        data = np.array([line.split() for line in chunks[1].splitlines()], dtype=float)
    else:
        if dtype == "float":
            data = np.frombuffer(chunks[1], np.single).reshape(nrow, ncol * ncomp)
        elif dtype == "double":
            data = np.frombuffer(chunks[1], np.double).reshape(nrow, ncol * ncomp)
        else:
            raise ValueError(f"Unsupported data type {dtype}")
    rdata = data[:, ::ncomp]
    gdata = data[:, 1::ncomp]
    bdata = data[:, 2::ncomp]
    if (len(bdata) != nrow) or (len(bdata[0]) != ncol):
        raise ValueError("Parsing matrix file failed")
    return rdata, gdata, bdata


def smx2nparray(data_str):
    """Convert Radiance sky matrix file to numpy array."""
    if not data_str.startswith(b"#?RADIANCE"):
        raise ValueError("No header found")
    if b"\r\n\r\n" in data_str:
        linesep2 = b"\r\n\r\n"
    else:
        linesep2 = b"\n\n"
    chunks = data_str.split(linesep2)
    header_str = chunks[0].decode()
    content = chunks[1:]
    nrow, ncol, ncomp, dtype = parsers.parse_rad_header(header_str)
    if dtype == "ascii":
        data = [i.splitlines() for i in content if i != b""]
        rdata = np.array(
            [[i.split()[::ncomp][0] for i in row] for row in data], dtype=float
        )
        gdata = np.array(
            [[i.split()[1::ncomp][0] for i in row] for row in data], dtype=float
        )
        bdata = np.array(
            [[i.split()[2::ncomp][0] for i in row] for row in data], dtype=float
        )
    else:
        if dtype == "float":
            data = np.frombuffer(b"\n\n".join(content), np.single).reshape(
                nrow, ncol * ncomp
            )
        elif dtype == "double":
            data = np.frombuffer(b"\n\n".join(content), np.double).reshape(
                nrow, ncol * ncomp
            )
        else:
            raise ValueError(f"Unsupported data type {dtype}")
        rdata = data[:, 0::ncomp]
        gdata = data[:, 1::ncomp]
        bdata = data[:, 2::ncomp]
    if ncol == 1:
        assert np.size(bdata, 1) == nrow
        assert np.size(bdata, 0) == ncol
        rdata = rdata.T
        gdata = gdata.T
        bdata = bdata.T
    else:
        assert np.size(bdata, 0) == nrow
        assert np.size(bdata, 1) == ncol
    return rdata, gdata, bdata


def numpy_mtxmult(mtxs, weight: Optional[Tuple[float, ...]] = None):
    """Matrix multiplication with Numpy."""
    weight = (47.4, 119.9, 11.6) if weight is None else weight
    resr = np.linalg.multi_dot([mat[0] for mat in mtxs]) * weight[0]
    resg = np.linalg.multi_dot([mat[1] for mat in mtxs]) * weight[1]
    resb = np.linalg.multi_dot([mat[2] for mat in mtxs]) * weight[2]
    return resr + resg + resb


def rad_mtxmult3(*mtxs, weights: tuple = (), no_header: bool = True):
    """Multiply matrices using dctimstep,
    Applying weights using rmtxop and remove header if needed.

    Args:
        mtx: Radiance matrices file paths
        weights: weight for the three channels
        no_header: whether to remove header from output

    Returns:
        output as byte strings

    Raises:
        Only works with four or two matrices.
    """
    if len(mtxs) not in (2, 4):
        raise ValueError("Only works with two or four matrices")
    if os.path.isfile(mtxs[-1]):
        cmd1 = ["dctimestep", "-od"] + list(mtxs)
        inp1 = None
    else:
        cmd1 = ["dctimestep", "-od"] + list(mtxs)[:-1]
        inp1 = mtxs[-1].encode()
    if weights == ():
        weights = (47.4, 119.9, 11.6)
        logger.warning("Using default photopic RGB weights")
    cmd2 = [
        "rmtxop",
        "-fa",
        "-c",
        str(weights[0]),
        str(weights[1]),
        str(weights[2]),
        "-",
        "-t",
    ]
    out1 = sp.Popen(cmd1, stdin=inp1, stdout=sp.PIPE)
    out2 = sp.Popen(cmd2, stdin=out1.stdout, stdout=sp.PIPE)
    if out1.stdout is not None:
        out1.stdout.close()
    if no_header:
        cmd3 = ["getinfo", "-"]
        out3 = sp.Popen(cmd3, stdin=out2.stdout, stdout=sp.PIPE)
        if out2.stdout is not None:
            out2.stdout.close()
        out = out3.communicate()[0]
    else:
        out = out2.communicate()[0]
    return out


def mtxmult(*mtxs):
    """Generic matrix multiplications."""
    if NUMPY_FOUND:

        def mtx_parser(fpath):
            if fpath.suffix == ".xml":
                proc = sp.run(
                    ["rmtxop", fpath],
                    check=True,
                    stdout=sp.PIPE,
                )
                raw = proc.stdout
            else:
                with open(fpath, "rb") as rdr:
                    raw = rdr.read()
            return mtxstr2nparray(raw)

        npmtx = [mtx_parser(mtx) for mtx in mtxs[:-1]]
        if os.path.isfile(mtxs[-1]):
            with open(mtxs[-1], "rb") as rdr:
                smx_str = rdr.read()
        else:
            smx_str = mtxs[-1]
        npmtx.append(smx2nparray(smx_str))
        return numpy_mtxmult(npmtx)
    return rad_mtxmult3(*mtxs)


def get_imgmult_cmd(*mtx: Path, odir: Path) -> List[str]:
    """Image-based matrix multiplication using dctimestep."""
    odir.mkdir(exist_ok=True)
    cmd = ["dctimestep", "-oc", "-o", str(odir / "%04d.hdr")]
    cmd += [str(m) for m in mtx]
    return cmd
